# NetWare-specific operations
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 

import os
import re
import shutil
import stat

from VMActions import VMActions
from exceptions import *
from log import log
from vmdisks import *
import mount
import util
try:
    import xnloader
except:
    pass

class VMNetWare(VMActions):
    def _kernel(self, base):
        return '%s/STARTUP/XNLOADER.SYS' % (base,)

    def getInstallImagesFromPath(self, options, path, progress=lambda f:None, supports_32nonpae=True):
        """Retrieves a Xen-enabled NetWare kernel and initrd.

        Retrieves a Xen-enabled NetWare kernel from the base path, and
        copies it to a temporary file.  Returns the (kernel, None) tuple of
        filenames."""
        kfd = open(self._kernel(path), 'r')
        try:
            kfn = util.copy_to_temp(kfd, prefix='netware.', progress=progress)
            xnloader.patch_netware_loader(kfn)
        finally:
            kfd.close()
        return (kfn, None)

    def getRuntimeImagesFromPath(self, root, os_type):
        """Returns a usable kernel and initrd pair found under 'root', or
        (None, None) if none are found."""
        kernel = '/nwserver/xnloader.sys'
        if os.path.exists(root + kernel):
            return (kernel, None)
        return (None, None)

    def setupOsSettings(self, guest, progress=lambda f:None):
        """Creates a temporary disk image, in which to pass the NetWare
        response file."""
        if not guest.options.os_settings:
            progress(1.0)
            return

        if guest.options.full_virt:
            # FIXME:  Need to support qemu's floppies...
            progress(1.0)
            return

        (fd, tmpdisk) = tempfile.mkstemp(prefix='netware-rsp.')
        try:
            disk = Disk(pdev=tmpdisk, ro=True, blocks=2880,
                        full_virt=guest.options.full_virt, vmname=guest.options.name,
                        disk_group=guest.options.disks)
            disk.setup(lambda f:progress(f*0.5))
            util.mkdosfsImage(tmpdisk, lambda f:progress(0.5+f*0.25))
            mounted_disk = mount.TemporaryLoopMount(tmpdisk)
            mountpoint = mounted_disk.mount(options='loop')
            progress(0.8)
            rf = False
            try:
                mode = os.stat(guest.options.os_settings).st_mode
                if stat.S_ISREG(mode):
                    shutil.copyfile(guest.options.os_settings, mountpoint + '/netware.rsp')
                    rf = True
                elif stat.S_ISDIR(mode):
                    rsp = re.compile('.*\.rsp$', re.IGNORECASE)
                    nlf = re.compile('.*\.nlf$', re.IGNORECASE)
                    nodes = os.listdir(guest.options.os_settings)
                    for node in nodes:
                        mode = os.stat('%s/%s' % (guest.options.os_settings, node)).st_mode
                        if not stat.S_ISREG(mode):
                            continue
                        if not rf and rsp.match(node):
                            rf = True
                            dst = 'netware.rsp'
                        elif nlf.match(node):
                            dst = node
                        else:
                            continue
                        log.debug("Copying  %s/%s -> %s/%s" % (guest.options.os_settings, node, mountpoint, dst))
                        shutil.copyfile('%s/%s' % (guest.options.os_settings, node),
                                        '%s/%s' % (mountpoint, dst))
                else:
                    raise ParamError(err=ParamError.E_INVALID, details='os-settings')
                progress(0.9)
            finally:
                mounted_disk.unmount()
                progress(1.0)
            if rf:
                guest.options._install_args = ' [INST: rf=a:netware.rsp]'
        except:
            os.unlink(tmpdisk)
            raise
        guest.tmpdisks.append(disk)

    def postInstall(self, guest):
        if guest.options.install and type(guest.options.source) is Disk:
            # Must remove the installation disk, otherwise the install will restart.
            i = guest.options.disks.index(guest.options.source)
            guest.options.disks.pop(i)
