# SUSE-specific operations
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 

from cStringIO import StringIO
import glob
import gzip
import os
import re
import shutil
import stat
import tempfile
import time
import urlgrabber.grabber as grabber

from VMLinux import VMLinux
from exceptions import *
from log import log
import caps
import mount
import msg
import util
import vmdisks
import vminstall

autoyast_vdev = vmdisks.get_possible_vdevs(False,False)[-1]

def die_zero(fn):
    try:
        if not os.stat(fn).st_size:
            raise RuntimeError
    except:
        try:
            os.unlink(fn)
        finally:
            raise RuntimeError

class VMSUSE(VMLinux):
    def _getPackageNames(self, base, fobj, supports_32nonpae=True):
        try:
            gobj = gzip.GzipFile(fileobj=StringIO(fobj.read()), mode='rb')
            lines = gobj.read().splitlines()

            arches = caps.compatible_arches(supports_32nonpae)
            assert(len(arches) > 0)
            for a in arches:
                # In sles11, the kernel packages have been broken up 
                # into kernel-xen, kernel-base, and kernel-extra.  kernel-base contains vmlinuz
                xenbase = True
                ext = a[1] + '-base'
                while True:
                    try:
                        packages = self._getPackageNamesByArch(base, lines, a[0], ext)
                        return a + packages
                    except InstSrcError, e:
                        log.debug("Failed to find valid install source for %s with kernel-%s" % (a[0], ext))
                        if xenbase is False:
                            break
                        ext = a[1]
                        xenbase = False
            raise e
        finally:
            fobj.close()

    def _getPackageNamesByArch(self, base, lines, arch, ext):
        kernel = 'kernel-' + ext
        log.debug("Searching for %s %s" % (arch, kernel))

        kernel_rpm = initrd_rpm = None
        kernel_disc = initrd_disc = None
        # The packages may or may not have a version.
        # There may or may not be a leading path component, specifying the disc.
        kre = re.compile('\.(/[a-zA-Z0-9]*)?/suse/%s/%s(-[0-9][-_\.0-9a-zA-Z]*)?\.rpm' % (arch, kernel))
        ire = re.compile('\.(/[a-zA-Z0-9]*)?/suse/%s/%s(-[0-9][-_\.0-9a-zA-Z]*)?\.rpm' % (arch, 'install-initrd'))
        for line in lines:
            m = kre.match(line)
            if m:
                log.debug("Found kernel line: '%s'" % (line,))
                paths = line.split('/')
                if len(paths) >= 5:
                    kernel_disc = paths[1]
                kernel_rpm = paths[-1]
                if initrd_rpm:
                    break
            else:
                m = ire.match(line)
                if m:
                    log.debug("Found initrd line: '%s'" % (line,))
                    paths = line.split('/')
                    if len(paths) >= 5:
                        initrd_disc = paths[1]
                    initrd_rpm = paths[-1]
                    if kernel_rpm:
                        break
        if not kernel_rpm or not initrd_rpm:
            raise InstSrcError(err=InstSrcError.E_BAD,
                               details=msg.suse_kernel_not_found)
        return [kernel_disc, kernel_rpm, initrd_disc, initrd_rpm]

    def _buildInitrd(self, krpm, irpm, progress=lambda f:None):
        (fd, ifn) = tempfile.mkstemp(prefix='inst-initrd.')
        tmpdir = tempfile.mkdtemp()
        log.info("Building inst-initrd...")
        try:
            os.system("cd -- '%s' && /usr/bin/rpm2cpio '%s' | /usr/bin/cpio -idum 2>/dev/null" % \
                (tmpdir, irpm))
            progress(0.2)
            os.system("%s/usr/sbin/mkinstallinitrd --kernel-rpm '%s' --libdir '%s' '%s'" % \
                (tmpdir, krpm, tmpdir + '/usr/lib/install-initrd', ifn))
            progress(0.9)
        finally:
            os.system("/bin/rm -rf -- '%s'" % (tmpdir,))
            os.unlink(irpm)
            progress(1.0)
        return ifn

    def _extractKernel(self, krpm, ext, progress=lambda f:None):
        (fd, kfn) = tempfile.mkstemp(prefix='vmlinuz-%s.' % ext)
        tmpdir = tempfile.mkdtemp()
        log.info("Extracting kernel from '%s' to '%s'" % (krpm, kfn))
        try:
            os.system("cd -- '%s' && /usr/bin/rpm2cpio '%s' | /usr/bin/cpio -idum -- '%s' 2>/dev/null" % \
                (tmpdir, krpm, './boot/vmlinuz-*-' + ext))
            progress(0.8)
            kernels = glob.glob('%s/boot/vmlinuz-*-%s' % (tmpdir, ext))
            for kernel in kernels:
                if not os.path.islink(kernel):
                    os.rename(kernel, kfn)
                    log.debug("Saving kernel '%s' -> '%s'" % (kernel, kfn))
                    break
        finally:
            os.system("/bin/rm -rf -- '%s'" % (tmpdir,))
            os.unlink(krpm)
            progress(1.0)
        return kfn

    def _getMediaRoots(self, root, _open=open):
        """Network install sources may be composed of multiple CDs.  YaST lists
        their location and ordering in ./yast/instorder.  Potentially any of
        these CDs may contain a PV kernel.
        """
        roots = ['.']
        try:
            fd = _open('%s/yast/instorder' % root)
            try:
                roots = fd.readlines()
                roots = map(lambda s: s.strip(), roots)
            finally:
                fd.close()
        finally:
            log.debug("Media roots: %s" % str(roots))
            return roots
        
    def getInstallImagesFromPath(self, options, root, progress=lambda f:None, supports_32nonpae=True):
        """Retrieves a Xen-enabled SUSE kernel and initrd from the base path, and
        copies them to temporary files.  Returns the (kernel, initrd) tuple of
        filenames."""
        searched = []
        arches = caps.compatible_arches(supports_32nonpae)
        for mediaroot in self._getMediaRoots(root):
            for a in arches:
                if a[0] == 's390x':
                    if options.os_type == "sles11" or options.os_type == "sled11":
                        kfn = '/%s/boot/%s/vmrdr.ikr' % (mediaroot, a[0])
                    else:
                        kfn = '/%s/boot/%s/linux' % (mediaroot, a[0])
                    ifn = '/%s/boot/%s/initrd' % (mediaroot, a[0])
                elif options.direct_kernel_boot:
                    if a[0] is not '':
                        k = '/boot/%s/loader/linux' % a[0]
                        kfn = '%s/%s%s' % (root, mediaroot, k)
                        ifn = '%s/%s/boot/%s/loader/initrd' % (root, mediaroot, a[0])
                    else:
                        k = '/boot'
                        kfn = '%s/%s%s/linux' % (root, mediaroot, k)
                        ifn = '%s/%s%s/initrd' % (root, mediaroot, k)
                    searched.append(k)
                else:
                    k = '/boot/%s/vmlinuz-%s' % (a[0], a[1])
                    searched.append(k)
                    kfn = '%s/%s%s' % (root, mediaroot, k)
                    ifn = '%s/%s/boot/%s/initrd-%s' % (root, mediaroot, a[0], a[1])
                log.debug("Checking for %s..." % kfn)
                if os.path.exists(kfn) and os.path.exists(ifn):
                    log.debug("Unpacked kernel/initrd found: %s %s" % (kfn, ifn))
                    kfd = ifd = None
                    try:
                        kfd = open(kfn, 'r')
                        ifd = open(ifn, 'r')
                        self.kfn = util.copy_to_temp(kfd, prefix='kernel.', progress=lambda f:progress(f*0.5))
                        try:
                            self.ifn = util.copy_to_temp(ifd, prefix='install-initrd.', progress=lambda f:progress(0.5+f*0.5))
                        except:
                            os.unlink(self.kfn)
                            raise
                    finally:
                        if kfd: kfd.close()
                        if ifd: ifd.close()
                    return (self.kfn, self.ifn)
        log.debug("No unpacked kernel/initrd found")
        index_path = root + '/INDEX.gz'
        searched.append('/INDEX.gz')
        log.debug("Path of INDEX.gz is '%s'" % (index_path,))
        try:
            fobj = open(index_path)
        except:
            raise InstSrcError(InstSrcError.E_BAD, details='\n'.join(searched))
        (arch, ext, kernel_disc, kernel_rpm, initrd_disc, initrd_rpm) = self._getPackageNames(root, fobj, supports_32nonpae)

        # FIXME:  may have to prompt to swap CDs...

        kfd = open('%s/suse/%s/%s' % (root, arch, kernel_rpm), 'r')
        krpm = util.copy_to_temp(kfd, prefix='kernel.', progress=lambda f:progress(f*0.1))
        kfd.close()

        ifd = open('%s/suse/%s/%s' % (root, arch, initrd_rpm), 'r')
        irpm = util.copy_to_temp(ifd, prefix='install-initrd.', progress=lambda f:progress(0.1+f*0.1))
        ifd.close()
        self.ifn = self._buildInitrd(krpm, irpm, lambda f:progress(0.2+f*0.4))
        self.kfn = self._extractKernel(krpm, ext, lambda f:progress(0.6+f*0.4))
        log.debug("Extracted installation kernel/initrd: %s %s" % (self.kfn, self.ifn))
        return (self.kfn, self.ifn)

    def getInstallImagesFromUrl(self, options, url, progress=lambda f:None, supports_32nonpae=True):
        """Retrieves a Xen-enabled SUSE kernel and initrd from the URL, and
        copies them to temporary files.  Returns the (kernel, initrd) tuple of
        filenames."""
        if not url:
            raise InstSrcError(err=InstSrcError.E_UNDEF)
        ftp = url.startswith('ftp://')
        http = url.startswith('http://')
        nfs = url.startswith('nfs://')
        smb = url.startswith('smb://')
        if nfs or smb:
            import urlparse
            # NOTE:  Python's urlparse module does not recognize nfs or smb,
            #        but the code handles it okay.  Add support:
            urlparse.uses_netloc.extend(['nfs', 'smb'])
            (_, host, path, _, _, _) = urlparse.urlparse(url)
            host = host.strip(':')
            if nfs:
                mnt = mount.TemporaryNFSMount(host, path)
            else:
                mnt = mount.TemporarySMBMount(host, path.lstrip('/'), options.smboptions)
            base = mnt.mount()
            urlopen = lambda path: open(path)
        elif ftp or http:
            base = url
            urlopen = lambda path: grabber.urlopen(path)
        else:
            # This should never happen (unless VMDefaults are out of sync?)
            raise InstSrcError(err=InstSrcError.E_UNSUPPORTED, details=url)
        progress(0.1)

        arches = caps.compatible_arches(supports_32nonpae)
        for mediaroot in self._getMediaRoots(base, urlopen):
            for a in arches:
                try:
                    if a[0] == 's390x':
                        if options.os_type == "sles11" or options.os_type == "sled11":
                            kfn = '/%s/boot/%s/vmrdr.ikr' % (mediaroot, a[0])
                        else:
                            kfn = '/%s/boot/%s/linux' % (mediaroot, a[0])
                        ifn = '/%s/boot/%s/initrd' % (mediaroot, a[0])
                    elif options.direct_kernel_boot:
                        kfn = '/%s/boot/%s/loader/linux' % (mediaroot, a[0])
                        ifn = '/%s/boot/%s/loader/initrd' % (mediaroot, a[0])
                    else:
                        kfn = '/%s/boot/%s/vmlinuz-%s' % (mediaroot, a[0], a[1])
                        ifn = '/%s/boot/%s/initrd-%s' % (mediaroot, a[0], a[1])
                    log.debug("Checking for %s..." % (base+kfn,))
                    kfd = urlopen(base + kfn)
                    self.kfn = util.copy_to_temp(kfd, prefix='kernel.', progress=lambda f:progress(0.1+f*0.45))
                    kfd.close()
                    die_zero(self.kfn)
                    ifd = urlopen(base + ifn)
                    self.ifn = util.copy_to_temp(ifd, prefix='install-initrd.', progress=lambda f:progress(0.55+f*0.45))
                    ifd.close()
                    die_zero(self.kfn)
                    log.debug("Unpacked kernel/initrd found: %s %s" % (kfn, ifn))
                    return (self.kfn, self.ifn)
                except:
                    continue
        log.debug("No unpacked kernel/initrd found")
        try:
            url = base + '/INDEX.gz'
            fobj = urlopen(url)
            (arch, ext, kernel_disc, kernel_rpm, initrd_disc, initrd_rpm) = self._getPackageNames(base, fobj, supports_32nonpae)

            # This is a heuristic:  If both are on DVD1 (or CD1, or foo1, ...)
            # or the disc isn't listed at all, then don't rewrite.  If the discs
            # are listed and at least one does _not_ end in '1', then assume
            # it's safe to rewrite the base path.  (The annoying corner-case
            # I'm trying to avoid is when both are on DVD1, but the url does
            # not actually end with "DVD1" -- not safe to rewrite!)
            # NOTE:  I can't back up above the NFS/SMB mount point, so (for
            #        now) no rewriting in those cases.
            if ((not nfs and not smb) and kernel_disc and initrd_disc and
                (not kernel_disc.endswith('1') or not initrd_disc.endswith('1'))):
                diskless = base.strip('/').split('/')[:-1]
                kbase = '/'.join(diskless + [kernel_disc])
                ibase = '/'.join(diskless + [initrd_disc])
            else:
                kbase = base
                ibase = base
            kurl = '/suse/%s/%s' % (arch, kernel_rpm)
            log.debug("URL of kernel is %s" % kurl)
            url = kbase + kurl
            kfd = urlopen(url)
            krpm = util.copy_to_temp(kfd, prefix='kernel.')
            kfd.close()
            progress(0.15)

            iurl = '/suse/%s/%s' % (arch, initrd_rpm)
            log.debug("URL of initrd is %s" % iurl)
            url = ibase + iurl
            ifd = urlopen(url)
            irpm = util.copy_to_temp(ifd, prefix='install-initrd.')
            ifd.close()
        except grabber.URLGrabError, e:
            raise InstSrcError(urlgrabber=e.errno, details='%s\n%s' % (url, e.strerror))
        except Error, e:
            raise
        except Exception, e:
            raise InstSrcError(err=InstSrcError.E_BAD, details='%s\n%s' % (url, str(e)))
        progress(0.2)
        self.ifn = self._buildInitrd(krpm, irpm, lambda f:progress(0.2+f*0.4))
        self.kfn = self._extractKernel(krpm, ext, lambda f:progress(0.6+f*0.4))
        log.debug("Extracted installation kernel/initrd: %s %s" % (self.kfn, self.ifn))
        return (self.kfn, self.ifn)

    def getInstallArgs(self, options):
        """Return SUSE's kernel arguments for installs."""
        global autoyast_vdev
        args = VMLinux.getInstallArgs(self, options)
        if options.source:
            if type(options.source) is vmdisks.Disk:
                if options.full_virt:
                    if options.source.get_vdevtype() == 'cdrom' and options.direct_kernel_boot:
                        src = 'cd:/'
                    else:
                        src = 'hd:/dev/%s' % options.source.get_vdev_str()
                else:
                    src = 'hd:/dev/%s' % options.source.get_vdev_str()
            else:
                src = options.source
            args += ' install=%s' % src
        ay = options.os_settings
        if ay:
            if ay.startswith('ftp://') or ay.startswith('tftp://') or \
               ay.startswith('http://') or ay.startswith('nfs://'):
                args += ' autoyast=%s' % (ay,)
            else:
                ay_vdev = autoyast_vdev
                if caps.is_xen() and options.os_type != 'sles11':
                    ay_vdev = 'xvdp'
                if options.direct_kernel_boot:
                    virtio = False
                    if caps.is_kvm():
                        if options.os_type == 'sles12' or options.os_type == 'sled12' or \
                           options.os_type == 'sles11' or options.os_type == 'sled11' or \
                           options.os_type == 'opensuse11' or options.os_type == 'opensuse12' or \
                           options.os_type == 'opensuse13':
                            virtio = True
                            autoyast_vdev = "vdb"
                        else:
                            autoyast_vdev = vmdisks.get_unused_vdevs(True, virtio, options.disks)[0]
                        ay_vdev = autoyast_vdev
                    else:
                        ay_vdev = vmdisks.get_possible_vdevs(True, virtio)[-1]
                # If a directory, look for either 'info' or 'autoinst.xml'
                if os.path.isdir(ay):
                    if os.path.isfile(ay + "/info"):
                        args += ' info=disk:/info'
                    if os.path.isfile(ay + "/autoinst.xml"):
                        args += ' autoyast=device://%s/autoinst.xml' % (ay_vdev,)
                else:
                    # File specified, look for either 'info' or anything else
                    if ay.endswith("info"):
                        args += ' info=disk:/info'
                    else:
                        args += ' autoyast=device://%s/%s' % (ay_vdev,os.path.basename(ay))
        return args

    def preprocessOptions(self, options):
        """Pulls settings out of extra_args."""
        if options.os_type == 'sles9' and options.full_virt:
            # This fixes the mouse tracking problem with FV guests
            #options.usb = True
            options.usb_mouse_abs = False
        if not options.extra_args:
            return
        args = options.extra_args.split()
        extra_args = ''
        for arg in args:
            if arg == 'TERM=xterm':
                # Discard.  Will be re-added later.
                pass
            elif arg.startswith('install='):
                if not options.source:
                    src = arg[8:]
                    if src.startswith('hd:/dev/'):
                        src = 'dev:/' + src[8:]
                    options.source = src
            elif arg.startswith('autoyast=') and not options.os_settings:
                options.os_settings = arg[9:]
            else:
                extra_args += arg + ' '
        options.extra_args = extra_args.strip()

    def setupOsSettings(self, guest, progress=lambda f:None):
        """Creates a temporary disk image, in which to pass the autoyast
        XML file."""
        global autoyast_vdev
        ay = guest.options.os_settings
        if not ay or ay.startswith('ftp://') or ay.startswith('tftp://') or \
                     ay.startswith('http://') or ay.startswith('nfs://'):
            pass
        elif not guest.options.full_virt or guest.options.direct_kernel_boot:
            (fd, tmpdisk) = tempfile.mkstemp(prefix='autoyast.')
            virtio = False
            if guest.options.direct_kernel_boot and caps.is_kvm():
                if guest.options.os_type == 'sles12' or \
                   guest.options.os_type == 'sled12' or \
                   guest.options.os_type == 'sles11' or \
                   guest.options.os_type == 'sled11' or \
                   guest.options.os_type == 'opensuse11' or \
                   guest.options.os_type == 'opensuse12' or \
                   guest.options.os_type == 'opensuse13':
                    virtio = True
                autoyast_vdev = vmdisks.get_unused_vdevs(guest.options.full_virt, virtio, guest.options.disks)[0]
                ay_vdev = autoyast_vdev
            else:
                ay_vdev = autoyast_vdev
                if not guest.options.full_virt and guest.options.os_type != 'sles11':
                    ay_vdev = 'xvdp'
                elif guest.options.direct_kernel_boot:
                    ay_vdev = 'hdd'
            disk = vmdisks.Disk(pdev=tmpdisk, vdev=ay_vdev, ro=True, blocks=2048,
                                full_virt=guest.options.full_virt, virtio=virtio,
                                vmname=guest.options.name)
            if guest.options.direct_kernel_boot and caps.is_kvm():
                # The autoyast file will use vdb.  Bump everything else down
                disks = guest.options.disks
                if len(disks) > 2 and disk.get_vdev_index() > 1:
                    for i in range(1, len(disks)):
                        disks[i].set_vdev(i+1)
                    autoyast_vdev = "vdb"
                    disk.set_vdev(autoyast_vdev)
            try:
                disk.setup(progress=lambda f:progress(f*0.5))
                util.mke2fsImage(tmpdisk, progress=lambda f:progress(0.5+f*0.25))
                mounted_disk = mount.TemporaryLoopMount(tmpdisk)
                mountpoint = mounted_disk.mount(options='loop')
                progress(0.8)
                try:
                    mode = os.stat(ay).st_mode
                    if stat.S_ISREG(mode):
                        if ay.endswith("info"):
                            shutil.copyfile(ay, mountpoint + '/info')
                        else:
                            shutil.copyfile(ay, mountpoint + "/" + os.path.basename(ay))
                    elif stat.S_ISDIR(mode):
                        if os.path.isfile(ay + "/autoinst.xml"):
                            shutil.copyfile(ay + '/autoinst.xml', mountpoint + '/autoinst.xml')
                        if os.path.isfile(ay + "/info"):
                            shutil.copyfile(ay + '/info', mountpoint + '/info')
                finally:
                    progress(0.9)
                    mounted_disk.unmount()
            except:
                os.unlink(tmpdisk)
                raise
            guest.tmpdisks.append(disk)
        progress(1.0)
