# Fully virtualized guest support
#
# Initially derived from virt-inst, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 

from exceptions import *
from full_virt_guest import FullVirtGuest
from xen_guest import XenGuest
from vmdisks import Disk
from log import log
import vminstall.advanced

class FullVirtXenGuest(FullVirtGuest, XenGuest):
    def __init__(self, options):
        XenGuest.__init__(self, options)

        self._cdrom = None
        self.features = {
            "acpi": True,
            "pae": True,  # required to install 64 bit!
        }
        self._qemu = '/usr/lib/xen/bin/qemu-system-i386'
        self.dkb = None

    def _get_os_xen(self):
        os = []
        if self.isInstall:
            if self.options.pxe_boot:
                bootdev = 'n'
            else:
                if type(self.options.source) is Disk:
                    if self.options.source.vdevtype == Disk.DEVICE_DISK:
                        bootdev = 'c'
                    else:
                        bootdev = 'd'
                else:
                    bootdev = 'd'
        else:
            bootdev = 'c'

        if self.isInstall and self.options.direct_kernel_boot and self.dkb:
            # Note these differences
            # loader = path to hvmloader
            # kernel = copied kernel from ISO/URL/CDROM
            # ramdisk = copied initrd
            # create another 'get' routine to return 'kernel', 'boot', 'loader', 'ramdisk'
            loader=self.options.loader
            ramdisk=self.options.initrd
            kernel=self.options.kernel
        else:
            kernel=self.options.loader
            boot=bootdev

    def _get_os_xml(self):
        os = []
        if self.isInstall:
            if self.options.pxe_boot:
                bootdev = 'network'
            elif type(self.options.source) is Disk:
                if self.options.source.vdevtype == Disk.DEVICE_DISK:
                    bootdev = 'hd'
                else:
                    bootdev = 'cdrom'
            else:
                bootdev = 'cdrom'
        else:
            bootdev = 'hd'

        os.append("<type>hvm</type>")
        if self.isInstall and self.options.direct_kernel_boot and self.dkb:
            os.append("<kernel>%s</kernel>" % self.dkb.kernel)
            os.append("<initrd>%s</initrd>" % self.dkb.initrd)
            extra = '%s %s' % (self.options.os_settings, self.options.extra_args)
            if type(self.options.source) is str:
                if self.options.source.startswith("ftp://") or \
                   self.options.source.startswith("http://") or \
                   self.options.source.startswith("tftp://") or \
                   self.options.source.startswith("nfs://"):
                    bootdev = None
                    extra = '%s %s' % (self.options._install_args, self.options.extra_args)
            os.append("<cmdline>%s</cmdline>" % extra)
        else:
            os.append("<loader>%s</loader>" % self.options.loader)
        if bootdev:
            os.append("<boot dev='%s'/>" % bootdev )
        return os

    def _get_features_xml(self):
        ret = []
        if self.options.apic:
            ret.append('<apic/>')
        for (k, v) in self.features.items():
            if v:
                ret.append('<%s/>' % k)
        return ret

    def _get_features_xen(self):
        ret = vminstall.advanced.get_advanced_features_xen(self.options)
        for (k, v) in self.features.items():
            if v:
                ret += '%s=1\n' %(k,)
        return ret

    def _get_mouse_xen(self):
        ret = ''
        if self.options.usb:
            #ret += "usb=1\n"
            if self.options.os_type != 'sles9':
                if self.options.usb_mouse_abs:
                    ret += "usbdevice='tablet'\n"
                else:
                    ret += "usbdevice='mouse'\n"
        return ret
        
    def GetConfigXml(self):
        return [ "<domain type='xen'>",
                   self.getCommonXml(),
                 [ '<os>',
                     self._get_os_xml(),
                   '</os>',
                   '<features>',
                     self._get_features_xml(),
                   '</features>',
                   '<devices>',
                   [ '<emulator>%s</emulator>' % self._qemu ],
                     self._get_disk_xml(),
                     self._get_network_xml(),
                     self._get_host_devices_xml(),
                     self._get_graphics_xml(),
                     self._get_sound_xml(),
                   '</devices>' ],
                 '</domain>' ]

    def GetConfigXen(self):
        if self.isInstall:
            if self.options.pxe_boot:
                bootdev = 'n'
            else:
                if self.options.direct_kernel_boot:
                    bootdev = 'c'
                elif type(self.options.source) is Disk:
                    if self.options.source.vdevtype == Disk.DEVICE_DISK:
                        bootdev = 'c'
                    else:
                        bootdev = 'd'
                else:
                    bootdev = 'd'
        else:
            bootdev = 'c'

        if self.isInstall and self.options.direct_kernel_boot and self.dkb:
            # Note these differences
            # loader = path to hvmloader
            # kernel = copied kernel from ISO/URL/CDROM
            # ramdisk = copied initrd
            # create another 'get' routine to return 'kernel', 'boot', 'loader', 'ramdisk'
            self.options.kernel = self.dkb.kernel
            self.options.initrd = self.dkb.initrd
            extra = '%s %s' % (self.options._install_args, self.options.extra_args)
            return """%(common)s
builder="hvm"
loader="%(loader)s"
kernel="%(kernel)s"
ramdisk="%(initrd)s"
boot="%(bootdev)s"
extra="%(extra)s"
%(disks)s
%(networks)s
%(host_devices)s
%(cpuid)s
%(cpucheck)s
%(graphics)s
%(sound)s
%(features)s
%(mouse)s
serial="pty"
                """ % { 'common': self.getCommonXen(), 'disks': self._get_disk_xen(),
                        'networks': self._get_network_xen(), 'host_devices': self._get_host_devices_xen(),
                        'cpuid': self._get_cpuid_xen(),
                        'cpucheck': self._get_cpu_check_xen(),
                        'graphics': self._get_graphics_xen(), 'sound': self._get_sound_xen(),
                        'features': self._get_features_xen(), 'bootdev': bootdev, 'extra': extra,
                        'loader': self.options.loader, 'kernel': self.options.kernel,
                        'initrd': self.options.initrd, 'mouse': self._get_mouse_xen()}

        return """%(common)s
builder="hvm"
boot="%(bootdev)s"
%(disks)s
%(networks)s
%(host_devices)s
%(cpuid)s
%(cpucheck)s
%(graphics)s
%(sound)s
%(features)s
%(mouse)s
serial="pty"
""" % { 'common': self.getCommonXen(), 'disks': self._get_disk_xen(),
        'networks': self._get_network_xen(), 'host_devices': self._get_host_devices_xen(),
        'cpuid': self._get_cpuid_xen(), 'cpucheck': self._get_cpu_check_xen(),
        'graphics': self._get_graphics_xen(), 'sound': self._get_sound_xen(),
        'features': self._get_features_xen(), 'bootdev': bootdev,
        'mouse': self._get_mouse_xen()}

    def PostInstallCheck(self):
        """Verify that the first hard disk is bootable."""
        hds = filter(lambda disk: disk.vdevtype == Disk.DEVICE_DISK, self.options.disks)
        if hds:
            vdev = hds[0].block_attach()
            try:
                if hds[0].has_bootsector():
                    return
            finally:
                hds[0].block_detach()
        raise VmError(VmError.E_NO_BOOTLOADER)

