################################################################################
# Fully virtualized guest support
#
# Initially derived from virt-inst, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 
################################################################################

from   console import Console
from   log     import log
import vminstall
import threading
import util
import caps
import time

import libvirt
from libvirt import VIR_DOMAIN_XML_SECURE

################################################################################
# CLASS: LibVirtConsole
#
# This class contains the libvirt-specific console implementation.
#
################################################################################
class LibVirtConsole(Console):

    #---------------------------------------------------------------------------
    # Class:  LibVirtConsole
    # Method: get_domid_from_uuid
    # Parms:
    #   uuid:   uuid from the domain
    #
    # This method gets the domain ID for the given UUID.
    #
    #---------------------------------------------------------------------------
    def get_domid_from_uuid(self, uuid):
        virConn = vminstall.hypervisor.connection.getconn()
        if virConn:
            virDom = None
            try:
                virDom = virConn.lookupByUUIDString(uuid)
            except:
                virConn = vminstall.hypervisor.connection.reset_connection()
                virDom = virConn.lookupByUUIDString(uuid)
            if virDom:
                domid = virDom.ID()
                log.debug("Domain %s has ID %d" % (uuid, domid))
                return domid
        log.debug("Domain %s is not running" % uuid)
        return None


    #---------------------------------------------------------------------------
    # Class:  LibVirtConsole
    # Method: get_tty_from_domid
    # Parms:
    #   domid:  domain ID
    #
    # This method gets the tty number for a given domain.
    #
    # This method is stubbed out in the base class.  It is expected that
    # anyone that subclasses this class will create their own method.
    #---------------------------------------------------------------------------
    def get_tty_from_domid(self, domid):
        if domid is None:
            return None
        if caps.is_xen():
            console_path = '/local/domain/%d/console/tty' % domid
            try:
                tty = util.run(['/usr/bin/xenstore-read', '-s', console_path])
            except:
                return None
            log.info('Domain ID %d has tty %s' % (domid, str(tty)))
            return tty
        # We can avoid xenstore entirely
        virConn = vminstall.hypervisor.connection.getconn()
        if virConn:
            virDom = virConn.lookupByID(domid)
            if virDom:
                xml = virDom.XMLDesc(VIR_DOMAIN_XML_SECURE)
                if xml:
                    tty = self.parse_xml_data(xml, 'console', 'tty')
        log.info('Domain ID %d has tty %s' % (domid, str(tty)))
        return tty


    #---------------------------------------------------------------------------
    # Class:  LibVirtConsole
    # Method: get_vnc_port_from_domid
    #
    # Parms:
    #   domid:      The domain ID of the domain that we are waiting to exit.
    #   timeout:    Wait only so long.
    #
    # This method finds the vnc port for the given domain.
    #
    # This method is stubbed out in the base class.  It is expected that
    # anyone that subclasses this class will create their own method.
    #---------------------------------------------------------------------------
    def get_vnc_port_from_domid(self, domid, timeout=0.0):
        if domid is None:
            return None
        ports = [None]
        virConn = vminstall.hypervisor.connection.getconn()
        if virConn:
            virDom = virConn.lookupByID(domid)
            if virDom:
                xml = virDom.XMLDesc(VIR_DOMAIN_XML_SECURE)
                if xml:
                    p = self.parse_xml_data(xml, 'graphics', 'port')
                    ports[0] = p
        port = ports[0]
        log.info('Domain ID %d has vnc port %s' % (domid, port))
        if port is not None:
            port = int(port)
        return port


    #---------------------------------------------------------------------------
    # Class:  LibVirtConsole
    # Method: wait_for_domid_to_exit
    #
    # Parms:
    #   domid:  The domain ID of the domain that we are waiting to exit.
    #   uuid:   UUID from the domain
    #
    # This method waits for the given domain to exit.
    #
    #---------------------------------------------------------------------------
    def wait_for_domid_to_exit(self, domid, uuid):
        """Wait for domid to exit (or, for domid's UUID to not match
        the specified uuid, which would indicate a race condition in
        which domid has already exited and was replace by another)."""
        if domid is None:
            return
        # Hack: Ugly polling method of waiting.  
        # Need a better way to know when the guest has exited
        log.debug("Polling libvirt for when guest goes away '%s'" % uuid)
        while True:
            try:
                virConn = vminstall.hypervisor.connection.getconn()
                if virConn:
                    virDom = virConn.lookupByID(domid)
                    if virDom:
                        u = virDom.UUIDString()
                        if u == uuid:
                            time.sleep(3.0)
                            continue
                break
            except:
                log.debug("Libvirt exception means guest has exited. '%s'" % uuid)
                break


