################################################################################
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 
#
# This is a class that hypervisors that have libvirt support can subclass.
#
################################################################################

from   hypervisor import Hypervisor
from   exceptions import ParamError
from   libvirt import VIR_DOMAIN_SHUTOFF
from   log import log 
import libvirt
import caps
import time
import dbus
import os
import hypervisor
import msg

#---------------------------------------------------------------------------
# Method: libvirt_error_handler
# Global handler to suppress the noise from libvirt
#---------------------------------------------------------------------------
def libvirt_error_handler(ctx, err):
    if err is not None:
        log.debug('Error: Libvirt - %s' % err)

def parse_string(uri):
    if uri.startswith("xen") or uri.startswith("qemu") or uri.startswith("test"):
        return LibVirtHypervisor(uri)
    raise ParamError(err=ParamError.E_INVALID, details=uri)

################################################################################
# Class: LibVirtHypervisor
#
# LibVirt Hypervisor specifics.
################################################################################
class LibVirtHypervisor(Hypervisor):
    """Thin abstraction to talk to libvirt.
    """
    XML_RPC_SOCKET = "/var/run/xend/xmlrpc.sock"

    #---------------------------------------------------------------------------
    # Method: __init__
    #
    # Comments:
    #   Sets the connection information to talk to libvirt.
    #
    def __init__(self, uri):
        self.uri = uri
        try:
            self.virConn = libvirt.open(self.uri)
        except:
            raise RuntimeError(msg.libvirt_connection)
        self.virNode = None
        self.virDom = None
        self.use_libvirt = True
        libvirt.registerErrorHandler(libvirt_error_handler, self.virConn)

    #---------------------------------------------------------------------------
    # Method: authenticate
    #---------------------------------------------------------------------------
    def _authenticate(self, virCreds, cbdata):
        if os.getuid() == 0:
            log.debug("Root credentials validated")
            return 0
        try:
            log.debug("Authenticate via policykit")
            bus = dbus.SessionBus()
            # Should work for sles11 and newer
            obj = bus.get_object("org.freedesktop.PolicyKit.AuthenticationAgent", "/")
            pkit = dbus.Interface(obj, "org.freedesktop.PolicyKit.AuthenticationAgent")
            pkit.ObtainAuthorization(virCreds[0][1], 0, os.getpid())
        except dbus.exceptions.DBusException, e:
            log.debug("Failed authentication using AuthenticationAgent, trying org.gnome.PolicyKit")
            if e.get_dbus_name() != "org.freedesktop.DBus.Error.ServiceUnknown":
                raise e
            # If PolicyKit < 0.7 try org.gnome.PolicyKit
            try:
                obj = bus.get_object("org.gnome.PolicyKit", "/org/gnome/PolicyKit/Manager")
                pkit = dbus.Interface(obj, "org.gnome.PolicyKit.Manager")
                pkit.ShowDialog(virCreds[0][1], 0)
            except Exception, e:
                raise e
        return 0

    #---------------------------------------------------------------------------
    # Method: delete
    #---------------------------------------------------------------------------
    def delete(self, domid):
        """Destroys a running instance; erases from libvirt"""
        try:
            if self.virDom is None:
                self.virDom = self.lookup(domid)
            if self.virDom:
                try:
                    self.virDom.destroy()
                except:
                    # May not be running
                    pass
                self.virDom.undefine()
        except:
            pass


    #---------------------------------------------------------------------------
    # Method: domain
    #---------------------------------------------------------------------------
    def domain(self, domid):
        try:
            self.virDom = self.lookup(domid)
            return self.virDom
        except:
            return None


    #---------------------------------------------------------------------------
    # Method: destroy
    #---------------------------------------------------------------------------
    def destroy(self, domid):
        """Destroys a running instance; does not erase from libvirt"""
        try:
            if self.virDom is None:
                self.virDom = self.lookup(domid)
            if self.virDom:
                self.virDom.shutdown()
        except:
            pass


    #---------------------------------------------------------------------------
    # Method: getconn
    #---------------------------------------------------------------------------
    def getconn(self):
        return self.virConn


    #---------------------------------------------------------------------------
    # Method: reset_connection
    #---------------------------------------------------------------------------
    def reset_connection(self, uri=None):
        if self.virConn is not None:
            uri = self.virConn.getURI()
            self.virConn.close()
            self.virConn = None
        if uri is None:
            hypervisor.connection = LibVirtHypervisor("qemu:///system")
        else:
            hypervisor.connection = LibVirtHypervisor(uri)
        return self.virConn


    #---------------------------------------------------------------------------
    # Method: get_dom0_min_memory
    #---------------------------------------------------------------------------
    def get_dom0_min_memory(self):
        # FIXME
        return 0


    #---------------------------------------------------------------------------
    # Method: domain_running
    #---------------------------------------------------------------------------
    def domain_running(self, domid):
        domain = self.domain(domid)
        if not domain:
            return False
        state = domain.state()
        if state[0] == VIR_DOMAIN_SHUTOFF:
            return False
        return True

    #---------------------------------------------------------------------------
    # Method: info
    #---------------------------------------------------------------------------
    def info(self):
        self.virNode = self.virConn.getInfo()
        if caps.is_xen():
            maxmem = caps.free_memory(True)
        else:
            maxmem = self.virNode[1]
        return dict([('nr_cpus', self.virNode[2]), ('total_memory', self.virNode[1]), ('max_hvm_memory', maxmem), ('max_para_memory', maxmem)])


    #---------------------------------------------------------------------------
    # Method: lookup
    #---------------------------------------------------------------------------
    def lookup(self, domid):
        """Lookup the domain by name, UUID, or hypervisor ID"""
        if type(domid) == int:
            virDom = self.virConn.lookupByID(domid)
        else:
            try:
                virDom = self.virConn.lookupByName(domid)
            except:
                try:
                    virDom = self.virConn.lookupByUUID(domid)
                except:
                    return None
        return virDom


    #---------------------------------------------------------------------------
    # Method: wait_domain
    #---------------------------------------------------------------------------
    def wait_domain(self, domid):
        #  This is ugly (polling instead of event driven) but XenAPI doesn't
        #  yet grok unix socket.
        #
        #  If domain and 'state' are both found, then assume VM is trying
        #  to come up, so keep waiting.  Otherwise, fail.
        for _ in xrange(10):
            domain = self.domain(domid)
            if not domain:
                return False
            wait = False
            if self.domain_running(domid):
                return True
            time.sleep(1.0)
        return False

