# Network Interface Card class
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 

import os
import string
import random
import caps
import util

from exceptions import *
import msg

################################################################################
# GLOBAL DATA
################################################################################
bridges = None


################################################################################
# GLOBAL FUNCTIONS
################################################################################

#-------------------------------------------------------------------------------
# Function: get_bridges
#
# If the bridges are already defined, simply return the bridges.
# If the bridges are not defined, then send the system command:
# "brctl show" which will dump the known bridges on the system.
# Parse the output and put just the bridge name in the bridge list.
#-------------------------------------------------------------------------------
def get_bridges(reload=False):
    global bridges

    if bridges is None or reload is True:
        bridges = []
        if caps.kvm_arch() == 's390x':
            rfd = os.popen("/sbin/ip link show")
            data = rfd.read()
            status = rfd.close()
            lines = data.splitlines()
            for line in lines:
                words = line.split()
                if len(words) > 1 and words[1].endswith(':'):
                    link = words[1].rstrip(':')
                    bridges.append(link)
        else:
            rfd = os.popen("/sbin/brctl show")
            data = rfd.read()
            status = rfd.close()
            lines = data.splitlines()[1:]
            for line in lines:
                if line[0] in (' ', '\t'):
                    continue
                words = line.split()
                bridges.append(words[0])
    return bridges

# s390x non-bridged networking
def get_interface_info(interface):
    is_up = False
    is_inet = False
    is_loopback = False

    def invoke_cmd(cmd):
        rfd = os.popen(cmd)
        data = rfd.read()
        status = rfd.close()
        lines = data.splitlines()
        return lines

    up_cmd = "/sbin/ip link show up " + interface
    lines = invoke_cmd(up_cmd)
    if lines:
        is_up = True 
    addr_cmd = "/sbin/ip addr show " + interface + " | grep \"inet \""
    lines = invoke_cmd(addr_cmd)
    if lines:
        is_inet = True 
    loopback_cmd = "/sbin/ip addr show " + interface + " | grep loopback"
    lines = invoke_cmd(loopback_cmd)
    if lines:
        is_loopback = True 

    return is_up, is_inet, is_loopback

# s390x non-bridged networking
def find_default_interface():
    bridge = None
    first_up = first_inet = None
    for b in bridges:
        if b == 'lo':
            continue
        macname = None
        ethname = None
        if b.find("@") != -1:
            ethname, ignore = b.split("@",1)
        else:
            ethname = b
        is_up, is_inet, is_lo = get_interface_info(ethname)
        if is_lo is True:
            continue
        if is_up is True and is_inet is True:
            bridge = b
            break
        elif first_up is None and is_up is True:
            first_up = b
        elif first_inet is None and is_inet is True:
            first_inet = b
    if bridge is None:
        if first_inet:
            bridge = first_inet
        elif first_up:
            bridge = first_up
        else:
            bridge = bridges[0]
    return bridge

#-------------------------------------------------------------------------------
# Function: parse_string
# Parms:
#   nic:    This is a comma separated string that contains the information
#           for the mac address, the bridge, and the model of a nic.
#           The format for the string is:
#               mac=<MAC ADDR>,bridge=<BRIDGE>,model=<MODEL>
#           Note that the parameters can be in any order.
#
# Invalid entries in the nic parameter will cause an exception.
# The input string is parsed and a Nic (class) is created.
#-------------------------------------------------------------------------------
def parse_string(nic):
    if nic == "":
        return XenNic()
    params = nic.split(',')
    mac = None
    bridge = None
    model = None
    for p in params:
        if p.startswith('mac='):
            mac = p[4:]
        elif p.startswith('bridge='):
            bridge = p[7:]
        elif p.startswith('model='):
            model = p[6:]
        else:
            raise ParamError(err=ParamError.E_INVALID, details=p)
    if caps.is_kvm() or caps.is_qemu():
        return KvmNic(mac=mac, bridge=bridge, model=model)
    else:
        return XenNic(mac=mac, bridge=bridge, model=model)


################################################################################
# Class: Nic
#
# Base class for all of the nic classes.
################################################################################
class Nic(object):

    ############################################################################
    # DATA
    ############################################################################

    models = {}                 # Create an empty dictionary representing the 
                                # valid NICs.


    ############################################################################
    # METHODS
    ############################################################################

    #---------------------------------------------------------------------------
    # Class: Nic
    # Method: __init__
    #---------------------------------------------------------------------------
    def __init__(self, number=None, mac=None, bridge=None, model=None, full_virt=True):
        self.number = number       # free-form, for caller's use
        self.mac = mac             # None or '11:22:33:44:55:66'
        if bridges is None:
            get_bridges()
        if bridge is None and bridges:
            bridge = bridges[0]
            if caps.kvm_arch() == 's390x':
                bridge = find_default_interface()
        self.bridge = bridge
        self.full_virt = full_virt
        self.model = model         # None or one of KvmNic or XenNic.models.keys()

    #
    # Property: mac
    #
    def get_mac(self):
        return self._mac
    def set_mac(self, mac):
        def badmac():
            raise ParamError(err=ParamError.E_INVALID, details=msg.mac_address)
        if mac is not None and type(mac) != str:
            badmac()
        if mac:
            octets = mac.split(':')
            if len(octets) != 6:
                badmac()
            for o in octets:
                if len(o) != 2 or o[0] not in string.hexdigits or o[1] not in string.hexdigits:
                    badmac()
        if mac:
            self._mac = mac.lower()
        else:
            self._mac = None
    mac = property(get_mac, set_mac)

    #
    # Property: bridge
    #
    def get_bridge(self):
        return self._bridge
    def set_bridge(self, bridge):
        self._bridge = bridge
    bridge = property(get_bridge, set_bridge)

    #
    # Property: model
    #
    # If the model is not set, then set the model as follows:
    #   fully-virtualized    -> Realtek 8139
    #   kvm para-virtualized -> virtio
    #   xen para-virtualized -> para
    # Otherwise, if the model is set, just return the current model.
    #
    def get_model(self):
        if self._model is None:
            if self.full_virt:
                # Only use virtio on platforms that have the driver.
                #if caps.is_kvm() or caps.is_qemu():
                #    return 'virtio'
                return 'rtl8139'
            else:
                return 'para'
        return self._model
    def set_model(self, model):
        if model is not None and model not in self.models.keys():
            raise ParamError(err=ParamError.E_INVALID, details='model')
        self._model = model
    model = property(get_model, set_model)


    #---------------------------------------------------------------------------
    # Class: Nic
    # Method: copy
    #
    # Creates a copy of this class.
    #---------------------------------------------------------------------------
    def copy(self):
        if caps.is_kvm() or caps.is_qemu():
            return KvmNic(number=self.number, mac=self.mac, bridge=self.bridge, 
                          model=self._model,  full_virt=self.full_virt)
        else:
            return XenNic(number=self.number, mac=self.mac, bridge=self.bridge, 
                          model=self._model,  full_virt=self.full_virt)


    #---------------------------------------------------------------------------
    # Class: Nic
    # Method: setup
    #
    # Setups up the bridges & mac address.
    #---------------------------------------------------------------------------
    def setup(self, progress=lambda f:None):
        if (caps.is_kvm() or caps.is_qemu()) and self.bridge is None:
            get_bridges()
            if bridges:
                self.bridge = bridges[0]
        if self.mac is None:
            mac = []
            # Courtesy of netif.py randomMAC() to avoid circular imports
            if caps.is_xen():
                mac = [ 0x00, 0x16, 0x3e, 
                        random.randint(0x00, 0x7f), random.randint(0x00, 0xff), random.randint(0x00, 0xff) ]
            else:
                mac = [ 0x52, 0x54, 0x00, 
                        random.randint(0x00, 0x7f), random.randint(0x00, 0xff), random.randint(0x00, 0xff) ]
            self.mac = ':'.join(map(lambda x: "%02x" % x, mac))

        progress(1.0)


    #---------------------------------------------------------------------------
    # Class: Nic
    # Method: update
    #
    # Verifies the parameters and sets the full_virt flag.
    #---------------------------------------------------------------------------
    def update(self, full_virt):
        if self._model is not None and self._model != 'para' and not full_virt:
            raise ParamError(err=ParamError.E_INVALID, details=str(self))
        self.full_virt = full_virt


    #---------------------------------------------------------------------------
    # Class: Nic
    # Method: __repr__
    #
    # Called by the repr built-in function and by string conversions
    # (reverse quotes) to compute the "official" string representation 
    # of an object.
    #
    # In the Nic class, it will be printed as a comma-separated list:
    #   <MAC ADDR>,<BRIDGE>,<MODEL>,<FULLY-VIRTUALIZED>
    #
    # If any of the parameters are blank, the comma-separator will 
    # still be in place.
    #---------------------------------------------------------------------------
    def __repr__(self):
        r = ''
        if self.number is not None:
            r += '%d:' % self.number
        mac = self.mac or ''
        r += '%s,%s,%s,%d' % (self.mac or '', self.bridge or '', self.model or '', self.full_virt)
        return r


    #---------------------------------------------------------------------------
    # Class: Nic
    # Method: __str__
    #
    # Called by the str built-in function and by the print statement to compute the
    # "informal" string representation of an object.  The return value must be a string
    # object.
    #
    # For the Nic class, the string representation will be as follows:
    #   <MODEL>;<MAC ADDR>
    #
    # If the mac address is empty, a random one will be generated.
    #---------------------------------------------------------------------------
    def __str__(self):
        s = '%s; ' % self.models[self.model]
        if self.mac is None:
            s += msg.random_mac
        else:
            s += self.mac
        return s


################################################################################
# Class: KvmNic
#
# KVM-specific NIC class.
################################################################################
class KvmNic(Nic):
    ############################################################################
    # DATA
    ############################################################################

    #
    # Define those NICs that are valid for a kvm environment.
    #
    models = {
        'e1000': msg.nic_e1000,
        'eepro100': msg.nic_eepro100,
        'i82551': "Intel 82551",        # Keeps crashing.
        'i82557b': "Intel 82557b",
        'i82559er': "Intel 82559er",
        'pcnet': msg.nic_pcnet,
        'ne2k_pci': msg.nic_ne2k_pci,
        'rtl8139': msg.nic_rtl8139,
        'virtio': "QEMU Virtualized NIC Card",
    }

    ############################################################################
    # METHODS
    ############################################################################

    #---------------------------------------------------------------------------
    # Class: KvmNic
    # Method: update
    #
    # Needed to define a separate one for kvm since the paravirtual driver
    # is difference than the one in xen.
    #---------------------------------------------------------------------------
    def update(self, full_virt):
        if self._model is not None and self._model != 'virtio' and not full_virt:
            raise ParamError(err=ParamError.E_INVALID, details=str(self))
        self.full_virt = full_virt


################################################################################
# Class: XenNic
#
# Xen-specific NIC class.
################################################################################
class XenNic(Nic):
    ############################################################################
    # DATA
    ############################################################################

    #
    # Define those NICs that are valid for a xen environment.
    #
    models = {
        'e100': msg.nic_e100,
        'e1000': msg.nic_e1000,
        'pcnet': msg.nic_pcnet,
        'ne2k_isa': msg.nic_ne2k_isa,
        'ne2k_pci': msg.nic_ne2k_pci,
        'rtl8139': msg.nic_rtl8139,
        'para': msg.paravirtualized
    }

