# Options scratch-area for VM settings
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 

import os
import string

from VMDefaults import VMDefaults
from exceptions import *
from globals import *
from log import log
from nics import XenNic
from nics import KvmNic
import caps
import msg
import util
import vmdisks
import vminstall
import hypervisor
import keytable
import cpuid
import paths
import advanced
import sound
import xml

on_exit = ["destroy", "restart", "preserve", "rename-restart", "coredump-destroy", "coredump-restart"]

class Options(dict):
    def __init__(self):
        default_keys = (
            'background',
            'debug',
            'disks',            # Disk class
            'extra_args',       # User's install or boot args
            'full_virt',
            'graphics',         # 'cirrus', 'none', 'para', 'vesa'
            'screen_resolution',
            'graphics_viewer',  # 'text', 'sdl', 'vnc', 'virt-viewer'
            'initrd',
            '_install_args',    # install-time args calculated by OS
            'install',
            'kernel',
            'keymap',
            'sound',            # 'all', 'es1370', sb16, 'ac97, 'ich6'
            'loader',
            'max_memoryMB',
            'memoryMB',
            'name',
            'description',      # description of this domain by user
            'nics',             # XenNic class
            'noautoconsole',
            'norestart',
            'on_crash',
            'on_poweroff',
            'on_reboot',
            'os_settings',
            'os_type',
            '_runtime_args',    # runtime-time args calculated by OS
            'source',
            'ui',
            'usb_mouse_abs',
            'uuid',
            'vcpus',
            'cpuid',
            'cpu_check',
            'vnc_password',
            'vncport',
            'pxe_boot',
            'pxe_addr',
            'pxe_label', 
            'direct_kernel_boot',
            'non_pae',
            'existing_os',
            'no_install',
            'preserve_on_error',
            'config_dir',
            'connect',
            'use_libvirt',
            'smboptions',
            'upgrade',          # Command line --upgrade
            'upgradeable',      # Supported operation for this OS?
            'upgrade_managed_vm', # Xenstore know about this vm
            'upgrade_config',   # config file or generated sxp file
            'passthrough',
            'host_devices',
            'use_xl',
            'multi_stage_install',
        )
        default_keys = default_keys + tuple(vminstall.advanced.get_advanced_args())
        dict.__init__(self, dict(map(lambda k: (k, None), default_keys)))
        self.__initialized = True

    def copy(self):
        dst = Options()
        for (key, value) in self.__dict__.items():
            if key == 'disks' and value is not None:
                dst.disks = []
                for disk in value:
                    if self.virtio is not None:
                        disk.virtio = self.virtio
                    elif self.os_type is not None:
                        disk.virtio = VMDefaults[self.os_type].virtio()
                    dst.disks.append(disk.copy())
            elif key == 'nics' and value is not None:
                dst.nics = []
                for nic in value:
                    dst.nics.append(nic.copy())
            elif key == 'host_devices' and value is not None:
                dst.host_devices = []
                for dev in value:
                    dst.host_devices.append(dev.copy())
            else:
                dst.__dict__[key] = value
        return dst

    def __getattr__(self, item):
        try:
            return self.__getitem__(item)
        except KeyError:
            raise AttributeError(item)

    def __setattr__(self, item, value):
        if not self.__dict__.has_key('_Options__initialized'):
            return dict.__setattr__(self, item, value)
        self.__setitem__(item, value)

    def __setitem__(self, item, value):
        """Sets an option, doing only basic order-independent validation
        such as type- and platform-validation."""

        log.debug("Setting options[%s] = %s" % (item, str(value)))

        # Raise AttributeError if attempting to add additional keys
        self.__getattr__(item)

        if value is None:
            self.__dict__[item] = None
            return

        def bad():
            raise ParamError(err=ParamError.E_INVALID, details=item)
        def check_type(v, types):
            if not type(v) in types:
                bad()

        if item == 'apic':
            check_type(value, (int, bool))
            value = bool(value)
        elif item == 'ui':
            if value not in ('gtk', 'text', 'yast'):
                bad()
        elif item == 'background':
            check_type(value, (bool,))
        elif item == 'debug':
            check_type(value, (int, bool))
            value = bool(value)
        elif item == 'disks':
            check_type(value, (list,))
            disks = value
        elif item == 'full_virt':
            check_type(value, (bool,))
            if value and not caps.is_hvm_capable():
                raise ArchResourceError(err=ArchResourceError.E_NO_VT)
        elif item == 'graphics':
            if value not in ('cirrus', 'none', 'para', 'vesa'):
                bad()
        elif item == 'screen_resolution':
            check_type(value, (str,))
        elif item == 'install':
            check_type(value, (int,bool))
        elif item == 'keymap':
            values = keytable.keytable.values()
            if value not in values:
                # FIXME, need real string
                #raise ParamError(err=ParamError.E_INVALID, details=msg.keymap)
                bad()
        elif item == 'sound':
            values = sound.get_sound_cards().keys()
            if value not in values:
                bad()
        elif item == 'loader':
            if not os.path.exists(value):
                # FIXME: the following bad was commented out, needs to be in.
                #bad()
                pass
        elif item == 'max_memoryMB':
            check_type(value, (int, float, str))
            value = int(value)
        elif item == 'memoryMB':
            check_type(value, (int, float, str))
            value = int(value)
        elif item == 'name':
            check_type(value, (str,))
            validchars = string.ascii_letters + string.digits + '_-.:+'
            for c in value:
                if not c in validchars:
                    raise ParamError(err=ParamError.E_INVALID, details=msg.name_alpha)
            if len(value) == 0 or len(value) > 255:
                raise ParamError(err=ParamError.E_INVALID, details=msg.name_len)
            #FIXME: the first time through here, the domain does not exist (correct behavior).
            # However, libvirt spits out "libvir: QEMU error : Domain not found"
            # It would be nice to remove that error message.
            if hypervisor.connection and hypervisor.connection.domain(value):
                if self.install != OS_INSTALL_UPGRADE:
                    raise VmError(err=VmError.E_CONF_EXISTS)
        elif item == 'nics':
            check_type(value, (list,))
            nics = value
        elif item == 'on_crash':
            if value not in on_exit:
                bad()
        elif item == 'on_poweroff' or item == 'on_reboot':
            on_exit_normal = on_exit[:]
            on_exit_normal.pop()
            on_exit_normal.pop()
            if value not in on_exit_normal:
                bad()
        elif item == 'os_settings':
            check_type(value, (str,))
        elif item == 'os_type':
            if not value in VMDefaults:
                bad()
        elif item == 'source':
            # str:  'ftp://foo/bar', 'dev:/xvda', etc
            # Disk:  source follows disk, even if disk gets renamed, unless
            #           somehow the disk becomes unbootable (e.g., second CD in FV)
            check_type(value, (str, vmdisks.Disk))
        elif item == 'usb':
            check_type(value, (int, bool))
        elif item == 'usb_mouse_abs':
            check_type(value, (bool,))
        elif item == 'uuid':
            if type(value) == str:
                # FIXME:  Check length, etc.
                pass
            elif type(value) == type(int):
                value = util.uuidToString(value)
            else:
                bad()
        elif item == 'vcpus':
            check_type(value, (int, float))
            value = int(value)
        elif item == 'cpuid':
            check_type(value, (list,))
            cpuid = value
        elif item == 'cpu_check':
            check_type(value, (list,))
            cpu_check = value
        elif item == 'pxe_boot':
            check_type(value, (int, bool))
            value = bool(value)
        elif item == 'pxe_addr':
            check_type(value, (str,))
        elif item == 'direct_kernel_boot':
            check_type(value, (int, bool))
            value = bool(value)
        elif item == 'non_pae':
            check_type(value, (int, bool))
            value = bool(value)
        elif item == 'upgrade':
            check_type(value, (int, bool))
            value = bool(value)
        elif item == 'upgrade_config':
            check_type(value, (str,))
        elif item == 'existing_os':
            check_type(value, (int, bool))
            value = bool(value)
        elif item == 'no_install':
            check_type(value, (int, bool))
            value = bool(value)
        elif item == 'preserve_on_error':
            check_type(value, (int, bool))
            value = bool(value)
        elif item == 'config_dir':
            if not os.path.exists(value):
                bad()
        elif item == 'passthrough':
            check_type(value, (int, bool))
            value = bool(value)
        elif item == 'host_devices':
            check_type(value, (list,))
            host_devices = value
        elif item == 'use_xl':
            check_type(value, (int, bool))
            value = bool(value)
        elif item == 'connect':
            import libvirt_hypervisor
            import xen_hypervisor
            import xl_hypervisor
            if not isinstance(value, libvirt_hypervisor.LibVirtHypervisor) and \
               not isinstance(value, xen_hypervisor.Xend) and \
               not isinstance(value, xl_hypervisor.Xl):
                bad()
        elif item == 'acpi' or item == 'blkif' or \
             item == 'hap' or item == 'hpet' or item == 'isa' or \
             item == 'localtime' or item == 'memory_sharing' or \
             item == 'monitor' or item == 'netif' or item == 'nomigrate' or \
             item == 'oos' or item == 'opengl' or item == 'pae' or \
             item == 'pci_msitranslate' or item == 'pci_power_mgmt' or \
             item == 'sdl' or item == 'superpages' or \
             item == 'suppress_spurious_page_faults' or item == 'timer_mode' or \
             item == 'tpmif' or item == 'viridian' or item == 'vnc' or \
             item == 'vpt_align' or item == 'xen_platform_pci' or item == 'virtio':
            check_type(value, (int, bool))
            value = bool(value)
        self.__dict__[item] = value

    def para_possible(self):
        if caps.is_xen():
            vmdefaults = VMDefaults[self.os_type]
            return 'xen' in vmdefaults.virt_support()
        else:
            return None

    def full_possible(self):
        vmdefaults = VMDefaults[self.os_type]
        return caps.is_hvm_capable() and ('hvm' in vmdefaults.virt_support() or
                                          'hvm-ext' in vmdefaults.virt_support())

    def full_preferred(self, user_request_full):
        vmdefaults = VMDefaults[self.os_type]
        supported = vmdefaults.virt_support()
        if self.para_possible():
            if self.full_possible():
                if user_request_full is not None:
                    return user_request_full
                return supported[0] in ('hvm', 'hvm-ext')
            else:
                if user_request_full:
                    if ('hvm' in supported or 'hvm-ext' in supported):
                        raise ArchResourceError(err=ArchResourceError.E_NO_VT)
                    else:
                        raise ParamError(err=ParamError.E_NOFULL)
                return False
        else:
            if self.full_possible():
                return True
            else:
                if ('hvm' in supported or 'hvm-ext' in supported):
                    raise ArchResourceError(err=ArchResourceError.E_NO_VT)
                else:
                    raise ParamError(err=ParamError.E_NOPARA)
            return True

def shallow_copy(src, dst):
    for key in src.__dict__.keys():
        value = getattr(src, key)
        dst.__dict__[key] = value

def calculate_max_memory(options, startup_mem, maximum_mem):
    # Returns the maximum initial memory and maximum total memory.
    # PV guests under about 1 Gig given too much maximum memory crash or hang.
    # HVM guests require about 2 Meg per 128 Meg total maximum memory.
    max_allowed = maximum_mem
    init_allowed = startup_mem
    try:
        hypervisor_info = hypervisor.connection.info()
    except:
        if options.os_type != 'pxe':
            max_allowed = VMDefaults[options.os_type].max_memory()
        return (init_allowed, max_allowed)
    if options.full_virt:
        free_mem = hypervisor_info['max_hvm_memory']
        if free_mem < startup_mem:
            startup_mem = free_mem
        while True:
            max_avail = (free_mem - startup_mem) + 2
            max_allowed = (max_avail / 2) * 128
            if maximum_mem < max_allowed:
                overhead = (maximum_mem / 128) * 2
            else:
                overhead = (max_allowed / 128) * 2
            if overhead < free_mem:
                init_allowed = free_mem - overhead
            else:
                init_allowed = free_mem - 4
            if max_allowed >= init_allowed:
                break
            startup_mem -= 32
    else:
        init_allowed = hypervisor_info['max_para_memory']
        if startup_mem < 2048:
            max_allowed = startup_mem / 0.0625
        else:
            max_allowed = VMDefaults[options.os_type].max_memory()
            if max_allowed < init_allowed:
                max_allowed = init_allowed
    return (init_allowed, max_allowed)

def calculate(options, defaults):
    """Re-calculates 'defaults' based on 'options'.  Error if 'options' is
    not self-consistent."""
    log.debug("start recalculation")

    # connect needs to be first
    if options.connect:
        defaults.connect = options.connect
        hypervisor.connection = options.connect
        options.use_libvirt = True
    elif hypervisor.connection is None:
        if not caps.is_xen() or not util.is_xend_running() and not options.use_xl:
            # NOTE: the KVM path is coming through here!  Yea!
            import libvirt_hypervisor
            if caps.is_xen():
                defaults.connect = libvirt_hypervisor.LibVirtHypervisor("xen:///")
            else:
                defaults.connect = libvirt_hypervisor.LibVirtHypervisor("qemu:///system")
            defaults.use_libvirt = options.use_libvirt = True
        else:
            import xen_hypervisor
            import xl_hypervisor
            if options.use_xl:
                defaults.connect = xl_hypervisor.Xl(None)
                defaults.use_xl = options.use_xl
            else:
                defaults.connect = xen_hypervisor.Xend(None)
            defaults.use_libvirt = options.use_libvirt = False
        hypervisor.connection = defaults.connect
    elif defaults.use_libvirt is None and not defaults.use_xl:
        defaults.use_libvirt = options.use_libvirt = hypervisor.connection.use_libvirt

    old_os_type = defaults.os_type
    if options.os_type:
        defaults.os_type = options.os_type
    else:
        options.os_type = defaults.os_type = default_os_type
    if old_os_type != 'pxe' and defaults.os_type == 'pxe':
        # PXE is treated like an os_type with a minimal set of default settings.
        # It really should just be an installation source, not an os type.
        # Copy the defaults for the type we are switching from so they can be
        # preserved when 'pxe' is selected as a source.
        excluded_keys = [ 'pxe_boot', 'pxe_addr', 'pxe_label', 'source', '_install_args' ]
        for key in defaults.__dict__.keys():
            if key not in excluded_keys:
                value = getattr(defaults, key)
                options.__dict__[key] = value
    vmdefaults = VMDefaults[defaults.os_type]

    if options.pxe_boot:
        log.debug("PXE boot specified")
        defaults.pxe_boot = options.pxe_boot
    else:
        defaults.pxe_boot = False

    if options.pxe_addr:
        defaults.pxe_addr = options.pxe_addr
    else:
        defaults.pxe_addr = None

    if options.full_virt:
        defaults.full_virt = options.full_virt
    else:
        defaults.full_virt = defaults.full_preferred(options.full_virt)

    if options.existing_os:
        defaults.existing_os = options.existing_os
        defaults.install = OS_INSTALL_NO
        if options.install is None:
            options.install = OS_INSTALL_NO

    if options.install is None:
        if options.upgrade:
            defaults.install = OS_INSTALL_UPGRADE
        else:
            defaults.install = OS_INSTALL_YES
        options.install = defaults.install
    else:
        defaults.install = options.install

    if options.upgrade_managed_vm:
        defaults.upgrade_managed_vm = options.upgrade_managed_vm
    else:
        defaults.upgrade_managed_vm = False

    #if defaults.install == OS_INSTALL_UPGRADE and not vmdefaults.is_upgradeable():
    #    raise ParamError(err=ParamError.E_INVALID, details=msg.exc_param_noupgrade)

    if options.upgrade:
        defaults.upgrade = True
        if options.upgrade_config:
            defaults.upgrade_config = options.upgrade_config
        else:
            if options.name and os.path.exists(options.name):
                defaults.upgrade_config = options.name
                if options.upgrade_config is None:
                    options.upgrade_config = options.name
                options.name = os.path.basename(options.name)
        if hypervisor.connection.domain(options.name):
            defaults.upgrade_managed_vm = True

    if options.name:
        if defaults.name != options.name:
            defaults.name = options.name
    elif defaults.name is None or old_os_type != defaults.os_type:
        i = 0
        while True:
            if defaults.os_type != 'pxe':
                name = defaults.os_type
            if i:
                name += '-%d' % i
            try:
                defaults.name = name
                break
            except VmError:
                i += 1
            except:
                defaults.name = None
                break

    if options.description:
        defaults.description = options.description

    defaults.non_pae = vmdefaults.supports_32nonpae()

    if options.no_install:
        defaults.no_install = options.no_install
    else:
        defaults.no_install = False

    if options.preserve_on_error:
        defaults.preserve_on_error = options.preserve_on_error
    elif vminstall.globals.default_preserve_on_error and not defaults.preserve_on_error:
        options.preserve_on_error = defaults.preserve_on_error = vminstall.globals.default_preserve_on_error
    else:
        defaults.preserve_on_error = False

    if options.memoryMB:
        initial_mem = options.memoryMB
    else:
        initial_mem = vmdefaults.default_memory()
    if initial_mem < vmdefaults.min_memory():
        initial_mem = vmdefaults.min_memory()
    if options.max_memoryMB:
        maximum_mem = options.max_memoryMB
    else:
        maximum_mem = vmdefaults.max_memory()
    if maximum_mem < initial_mem:
        maximum_mem = initial_mem

    max_init_allowed, max_allowed = calculate_max_memory(options, initial_mem, maximum_mem)

    if initial_mem > max_init_allowed:
        initial_mem = max_init_allowed
    defaults.memoryMB = initial_mem
    if maximum_mem > max_allowed:
        maximum_mem = max_allowed
    defaults.max_memoryMB = maximum_mem 

    if options.vcpus:
        defaults.vcpus = options.vcpus
    else:
        defaults.vcpus = vmdefaults.num_vcpus()

    if not options.cpuid:
        defaults.cpuid = []
    else:
        defaults.cpuid = options.cpuid

    if not options.cpu_check:
        defaults.cpu_check = []
    else:
        defaults.cpu_check = options.cpu_check

    if options.loader:
        defaults.loader = options.loader
    else:
        if defaults.full_virt:
            defaults.loader = '/usr/lib/xen/boot/hvmloader'
        else:
            defaults.loader = 'pygrub'

    # The command line can take a UUID, but the TUI or GUI won't let the
    # user specify one.  So, if one didn't exist on the command line,
    # stick with the first randomly generated one.
    if options.uuid is None:
        options.uuid = util.uuidToString(util.randomUUID())
    defaults.uuid = options.uuid

    if options.on_crash:
        defaults.on_crash = options.on_crash
    else:
        defaults.on_crash = 'destroy'

    if options.on_poweroff:
        defaults.on_poweroff = options.on_poweroff
    else:
        defaults.on_poweroff = 'destroy'

    if options.on_reboot:
        defaults.on_reboot = options.on_reboot
    else:
        defaults.on_reboot = 'restart'

    if options.keymap:
        defaults.keymap = options.keymap
    else:
        defaults.keymap = None

    if options.sound:
        defaults.sound = options.sound
    else:
        defaults.sound = vmdefaults.sound()

    if options.graphics_viewer == 'sdl':
        defaults.graphics_viewer = 'sdl'
    elif options.graphics_viewer == 'vnc':
        defaults.graphics_viewer = 'vnc'
        defaults.vnc_password = options.vnc_password or ''
        # Nasty hack due to the fact that xen's config file has no escaping
        # FIXME:  drop support for xen's config file; move to XenAPI or libvirt
        if "'" in defaults.vnc_password or "," in defaults.vnc_password:
            raise ParamError(err=ParamError.E_INVALID, details='vnc_password')
        defaults.vncport = options.vncport
    elif vminstall.globals.default_graphics_viewer and not defaults.graphics_viewer:
        options.graphics_viewer = defaults.graphics_viewer = vminstall.globals.default_graphics_viewer
    else:
        if os.path.exists('/usr/bin/virt-viewer') and options.use_libvirt is True:
            defaults.graphics_viewer = 'virt-viewer'
        else:
            defaults.graphics_viewer = 'vnc'
            defaults.vncport = options.vncport
            defaults.vnc_password = options.vnc_password or ''
            if "'" in defaults.vnc_password or "," in defaults.vnc_password:
                raise ParamError(err=ParamError.E_INVALID, details='vnc_password')

    if options.graphics == 'none':
        defaults.graphics = 'none'
    elif defaults.full_virt:
        if options.graphics in ('cirrus', 'vesa'):
            defaults.graphics = options.graphics
        else:
            if defaults.os_type == 'oes2nw':
                defaults.graphics = 'vesa'
            elif caps.kvm_arch() == 's390x':
                defaults.graphics = 'none'
            else:
                defaults.graphics = 'cirrus'
    else:
        # Rhel7 kernel doesn't have framebuffer device built in kernel
        if options.graphics is None and options.os_type == "rhel7":
            options.graphics = defaults.graphics = 'none'
        else:
            defaults.graphics = 'para'

    if options.screen_resolution:
        defaults.screen_resolution = options.screen_resolution

    if options.disks is not None:
        defaults.disks = options.disks
    else:
        if defaults.install == OS_INSTALL_YES:
            defaults.disks = vmdefaults.disks()
        else:
            defaults.disks = []  # Not installing; user must specify existing disk
    for d in defaults.disks:
        d.update(full_virt=defaults.full_virt, vmname=defaults.name)
        d.disk_group = defaults.disks
    if len(defaults.disks):
        disks = defaults.disks
        log.debug('Checking for conflicts between disks: %s' % map(str, disks))
        for i in range(len(disks)-1):
            for j in range(i+1, len(disks)):
                if (disks[i].vdev == disks[j].vdev or
                    disks[i].pdev == disks[j].pdev):
                    raise DiskResourceError(err=DiskResourceError.E_CONFLICT,
                                            details='%s\n%s' % (disks[i], disks[j]))
        defaults.disks.sort(lambda d1,d2: cmp(d1.vdev, d2.vdev))
        if not defaults.disks[0].args.has_key('sparse'):
            defaults.disks[0].args['sparse'] = True

    if options.nics is not None:
        defaults.nics = options.nics
    else:
        defaults.nics = vmdefaults.nics()
    nics = defaults.nics
    if len(nics):
        for i in range(len(nics)-1):
            for j in range(i+1, len(nics)):
                if (nics[i].mac is not None and
                    nics[i].mac == nics[j].mac):
                    raise ParamError(err=ParamError.E_INVALID, details=msg.mac_address)
    for n in nics:
        n.update(defaults.full_virt)

    if options.passthrough is not None:
        defaults.passthrough = options.passthrough
    else:
        defaults.passthrough = None
    if options.host_devices is not None:
        defaults.host_devices = options.host_devices

    if options.extra_args:
        defaults.extra_args = options.extra_args
    else:
        defaults.extra_args = ''

    defaults.source = None
    if vminstall.globals.default_install_source and not options.source and not options.pxe_boot:
        if vminstall.globals.default_install_source == 'pxe':
            options.pxe_boot = defaults.pxe_boot = True
        else:
            options.source = vminstall.globals.default_install_source
    if options.source and not options.pxe_boot:
        defaults.source = options.source
        # FIXME:  full vs para affects what protocols are supported!
        source = defaults.source
        if type(source) is vmdisks.Disk:
            source = 'dev:/'
        supported = False
        protocols = vmdefaults.install_sources()[1]
        for protocol in protocols:
            if source.startswith(protocol):
                supported = True
                break
        if not supported:
            # If we switch os types and source is incompatable with new type, just clear it
            options.source = defaults.source = None
        if type(defaults.source) is str and defaults.source.startswith('dev:/'):
            # User specified string; convert to disk.
            vdev = defaults.source[5:]
            for disk in defaults.disks:
                if str(disk.get_vdev_index()) == vdev or disk.get_vdev_str() == vdev:
                    defaults.source = disk
                    break
        if type(defaults.source) is vmdisks.Disk and defaults.source not in defaults.disks:
            log.debug("source disk has been removed; re-defaulting source")
            options.source = defaults.source = None
        if defaults.full_virt and type(defaults.source) is vmdisks.Disk and \
            defaults.source not in vmdisks.get_fv_boot_disks(defaults.disks):
            log.debug("source disk not valid to boot from for FV; re-defaulting source")
            options.source = defaults.source = None
    if defaults.source is None:
        if defaults.full_virt:
            fv_boot_disks = vmdisks.get_fv_boot_disks(defaults.disks)
            log.debug("possible FV boot disks: %s" % map(str, fv_boot_disks))
            if fv_boot_disks:
                defaults.source = fv_boot_disks[0]
            elif caps.is_kvm():
                dom0_inst_source = util.getDom0InstallSource()
                if dom0_inst_source:
                    if dom0_inst_source.startswith('ftp:') or \
                       dom0_inst_source.startswith('http:') or \
                       dom0_inst_source.startswith('smb:') or \
                       dom0_inst_source.startswith('nfs:'):
                        options.source = defaults.source = dom0_inst_source
        elif 'dev:/' in vmdefaults.install_sources()[1]:
            vdevs = filter(lambda d: d.exists(), defaults.disks)
            if vdevs:
                # Slight hack, to prefer cdroms over disks and floppies.
                vdevs.sort(lambda x,y: cmp(x._vdevType, y._vdevType))
                defaults.source = vdevs[0]

    if options.os_settings:
        defaults.os_settings = options.os_settings
        if defaults.os_settings and vmdefaults.auto_install() is None:
            raise ParamError(err=ParamError.E_NOAUTO)
    else:
        defaults.os_settings = ''

    defaults.vncport = options.vncport
    defaults.norestart = options.norestart
    defaults.debug = options.debug
    defaults.ui = options.ui
    defaults.background = options.background
    if options.use_xl:
        if defaults.graphics_viewer is None or defaults.graphics_viewer == 'virt-viewer':
            options.graphics_viewer = defaults.graphics_viewer = 'vnc'
    defaults.use_xl = options.use_xl
    defaults.multi_stage_install = options.multi_stage_install = vmdefaults.multi_stage_install()

    if defaults.background:
        defaults.noautoconsole = True
    else:
        defaults.noautoconsole = options.noautoconsole

    # User can't specify _install_args or _runtime_args,
    # but the OS-specific code may have.
    defaults._install_args = vmdefaults.actions().getInstallArgs(defaults)
    if options._install_args:
        defaults._install_args += ' %s' % options._install_args
    defaults._runtime_args = vmdefaults.actions().getRuntimeArgs(defaults)
    if options._runtime_args:
        defaults._runtime_args += ' %s' % options._runtime_args

    if options.config_dir:
        defaults.config_dir = options.config_dir
    else:
        defaults.config_dir = paths.configdir

    if options.smboptions:
        defaults.smboptions = options.smboptions

    if options.direct_kernel_boot is not None:
        if vmdefaults.supports_direct_kernel_boot() and options.direct_kernel_boot is True:
            defaults.direct_kernel_boot = True
        else:
            defaults.direct_kernel_boot = False
    else:
        if defaults.full_virt and vmdefaults.supports_direct_kernel_boot():
            if type(defaults.source) is vmdisks.Disk and caps.is_xen():
                defaults.direct_kernel_boot = False
            else:
                defaults.direct_kernel_boot = vmdefaults.direct_kernel_boot()
        else:
            options.direct_kernel_boot = False
            defaults.direct_kernel_boot = False

    # Advanced options shared
    if options.apic is not None:
        defaults.apic = options.apic
    else:
        defaults.apic = vmdefaults.apic()
    defaults.acpi = options.acpi
    if options.hpet is not None:
        defaults.hpet = options.hpet
    else:
        defaults.hpet = vmdefaults.hpet()
    if options.usb is not None:
        defaults.usb = options.usb
    else:
        defaults.usb = vmdefaults.usb()
    defaults.pae = options.pae
    # Advanced options Xen specific
    if caps.is_xen():
        defaults.blkif = options.blkif
        defaults.hap = options.hap
        defaults.isa = options.isa
        defaults.localtime = options.localtime
        defaults.memory_sharing = options.memory_sharing
        defaults.monitor = options.monitor
        defaults.netif = options.netif
        defaults.nomigrate = options.nomigrate
        defaults.oos = options.oos
        defaults.opengl = options.opengl
        defaults.pci_msitranslate = options.pci_msitranslate
        defaults.pci_power_mgmt = options.pci_power_mgmt
        defaults.sdl = options.sdl
        defaults.superpages = options.superpages
        defaults.suppress_spurious_page_faults = options.suppress_spurious_page_faults
        defaults.timer_mode = options.timer_mode
        defaults.tpmif = options.tpmif
        defaults.viridian = options.viridian
        defaults.vnc = options.vnc
        defaults.vpt_align = options.vpt_align
        defaults.xen_platform_pci = options.xen_platform_pci
    else:
        if options.virtio is not None:
            defaults.virtio = options.virtio
        else:
            defaults.virtio = vmdefaults.virtio()
    
    # Now give the OS a chance to voice its opinion:
    vmdefaults.actions().preprocessOptions(defaults)

    log.debug("end recalculation")

def initialize_vm_from_config_file(options, vm_name):
    log.debug('Initializing vm-install internal data from configuration file, %s' % vm_name)
    if not options.upgrade_managed_vm:
        vm_name = options.upgrade_config
    (fname, lines, bname) = util.read_config_file(vm_name, options.upgrade_managed_vm, 0)
    log.debug('Returned from reading configuration file: %s' % fname)
    if fname is None:
        log.debug('Error reading configuration file for %s' % vm_name)
        raise InstSrcError(err=InstSrcError.E_CONFIG)
    if not util.is_xend_running():
        xml.merge(fname, options)
        os.unlink(fname)
        return
    os.unlink(fname)
    extra_args = options.extra_args
    options.disks = None
    options.nics = None
    image_start = 0
    image_data = []
    device_start = 0
    device_data = []
    for line in lines:
        if options.upgrade_managed_vm:
            # This is an sxp file format
            line = line.strip(" ")
            if line.startswith("(image"):
                image_start = 1
                image_data = [line]
                continue
            if line.startswith("(device"):
                device_start = 1
                device_data = [line]
                continue
            if image_start > 0:
                image_data.append(line)
                if line.startswith("("):
                    image_start += 1
                if line.endswith(")\n"):
                    image_start -= 1
                if image_start:
                    continue
            if device_start > 0:
                device_data.append(line)
                if line.startswith('('):
                    device_start += 1
                if line.endswith(')\n'):
                    device_start -= 1
                if device_start:
                    continue
            if len(image_data):
                kernel = util.lookup_sxp_value(image_data, 'kernel')
                if kernel:
                    options.kernel = kernel
                initrd = util.lookup_sxp_value(image_data, 'ramdisk')
                if initrd:
                    options.initrd = initrd
                options.extra_args = util.lookup_sxp_value(image_data, 'args')
                if extra_args:
                    if not options.extra_args:
                        options.extra_args = extra_args
                    else:
                        options.extra_args = extra_args + ' ' + options.extra_args
                image_data = ""
            elif len(device_data):
                if util.lookup_sxp_value(device_data, 'vif'):
                    mac = util.lookup_sxp_value(device_data, 'mac')
                    bridge = util.lookup_sxp_value(device_data, 'bridge')
                    model = util.lookup_sxp_value(device_data, 'model')
                    nic = XenNic(mac=mac, bridge=bridge, model=model)
                elif util.lookup_sxp_value(device_data, 'vbd') or util.lookup_sxp_value(device_data, 'tap'):
                    uname = util.lookup_sxp_value(device_data, 'uname')
                    dev = util.lookup_sxp_value(device_data, 'dev ')
                    dev = dev.split(":")
                    disktype = 'disk'
                    if len(dev) > 1:
                        disktype = dev[1]
                    dev = dev[0]
                    mode = util.lookup_sxp_value(device_data, 'mode')
                    disk_str = uname + ',' + dev + ',' + disktype + ',' + mode
                    disk = vmdisks.disks.parse_string(disk_str, False)
                    if options.disks:
                        options.disks.append(disk)
                    else:
                        options.disks = [disk]
                device_data = ""
            line = line.strip(" )(\n")
            value = line.split(" ")
            if len(value):
                vname = value[0]
            if len(value) < 2:
                continue
            value = value[1].strip('\' ')
        else:
            # This is a legacy config file format
            value = line[:-1].split('=', 1)[1]
            value = value.strip('"')
            vname = line[:-1].split('=', 1)[0]
            vname = vname.strip(" ")
        if vname == 'uuid':
            options.uuid = value
        elif vname == 'memory':
            options.memoryMB = int(value)
        elif vname == 'maxmem':
            options.max_memoryMB = int(value)
        elif vname == 'vcpus':
            options.vcpus = int(value)
        elif vname == 'on_poweroff':
            options.on_poweroff = value
        elif vname == 'on_reboot':
            options.on_reboot = value
        elif vname == 'on_crash':
            options.on_crash = value
        elif vname == 'builder':
            if value == 'linux':
                options.full_virt = False
            else:
                options.full_virt = True
        elif vname == 'bootloader' or vname == 'kernel':
            options.loader = value
        elif vname == 'bootargs' or vname == 'bootloader_args':
            substr = value.split(',')
            if len(substr) > 1:
                options.initrd = substr[1]
            substr = substr[0].split(':')
            if len(substr) > 1:
                options.kernel = substr[1]
        elif vname == 'extra':
            if extra_args:
                options.extra_args = extra_args + ' ' + value
            else:
                options.extra_args = value
        elif vname == 'disk':
            value = value.strip('[]')
            config_disks = value.split('\',')
            if options.disks:
                del options.disks[0:len(options.disks)]
            for cdisk in config_disks:
                cdisk = cdisk.strip(' \'')
                cdisk_parts = cdisk.split(',')
                if cdisk_parts[0].endswith('/dev/cdrom') or cdisk_parts[0].endswith('/dev/sr0'):
                    type='cdrom'
                else:
                    type='disk'
                item=0
                disk_str = ""
                for d in cdisk_parts:
                    type_check = d.split(":")
                    if len(type_check) > 1 and type_check[1] == 'cdrom':
                        type='cdrom'
                        d = type_check[0]
                    disk_str += d
                    item += 1
                    if item < len(cdisk_parts):
                        disk_str += ','
                    if item == 2:
                        disk_str += type + ','
                if disk_str:
                    disk = vmdisks.disks.parse_string(disk_str, False)
                    if options.disks:
                        options.disks.append(disk)
                    else:
                        options.disks = [disk]
        elif vname == 'vif':
            value = value.strip('[]')
            config_nics = value.split('\',')
            if options.nics:
                del options.nics[0:len(options.nics)]
            for cnic in config_nics:
                cnic = cnic.strip(' \'')
                nic = vminstall.nics.parse_string(cnic)
                if options.nics:
                    options.nics.append(nic)
                else:
                    options.nics = [nic]
    log.debug('Internal data initialization from configuration file complete')

def finalize(options):
    """Checks the options for conflicts or incomplete options."""
    # The Options class will guarentee that no single setting is invalid,
    # but since it has no concept of ordering or completeness, it can't
    # check for missing values, conflicts, etc.

    vmdefaults = VMDefaults[options.os_type]

    if options.name is None:
        log.debug("validation failed: options.name is None")
        raise ParamError(err=ParamError.E_INVALID, details=msg.name_len)

    if len(options.disks) == 0:
        raise DiskResourceError(err=DiskResourceError.E_NOT_FOUND)

    for disk in options.disks:
        if not disk.validate():
            if options.ui == 'text':
                vminstall.text.interface.output(vminstall.msg.warning + ': ' + vminstall.msg.exc_disk_overcommit + '\n')
                if not vminstall.text.interface.prompt_yes_no('', True):
                    raise DiskResourceError(DiskResourceError.E_NO_SPACE, details=disk.get_pdev())

    for nic in options.nics:
        if not options.full_virt and nic.model != 'para':
            raise ParamError(err=ParamError.E_INVALID, details=eval('msg.nic_' % nic.model))
        nic.setup()

    init_memory, max_memory = calculate_max_memory(options, options.memoryMB, options.max_memoryMB)
    if options.memoryMB > init_memory:
        raise MemoryResourceError(err=MemoryResourceError.E_NO_MEMORY)

    if options.full_virt:
        if options.install == OS_INSTALL_YES:
            cdrom = None
            for disk in options.disks:
                if disk._vdevType == vmdisks.Disk.DEVICE_CDROM:
                    cdrom = disk
                    break
            if cdrom is None and options.pxe_boot is False and options.direct_kernel_boot is False:
                log.debug("validation failed: No CDROM found for FV VM")
                raise InstSrcError(err=InstSrcError.E_CDROM)

        if options.graphics not in ('cirrus', 'none', 'vesa'):
            raise ParamError(err=ParamError.E_INVALID, details='graphics')

        # Work-around for Qemu slowness (bug #238458):  Virtual CD-ROMs
        # slow to a crawl if they share the virtual IDE controller with
        # a hard disk.  Attempt to push the CD-ROM(s) to their own
        # controller, without changing the ordering, if possible.
        if (len(filter(lambda d: d.vdevtype == vmdisks.Disk.DEVICE_CDROM, options.disks)) > 0):
            moved_index = True
            if len(options.disks) == 2:
                (options.disks[0].vdev, options.disks[1].vdev) = (0, 2)
            elif len(options.disks) == 3:
                if map(lambda d: d.vdevtype, options.disks).index(vmdisks.Disk.DEVICE_CDROM) == 2:
                    (options.disks[0].vdev, options.disks[1].vdev, options.disks[2].vdev) = (0, 1, 2)
                else:
                    (options.disks[0].vdev, options.disks[1].vdev, options.disks[2].vdev) = (0, 2, 3)
            else:
                moved = False
            if moved_index:
                options._install_args = vmdefaults.actions().getInstallArgs(options)
            log.debug('Re-distributed disks: %s' % map(str, options.disks))
    else:
        if options.install == OS_INSTALL_YES:
            if not options.pxe_boot and not options.source:
                log.debug("validation failed: No install source found for PV VM.")
                raise InstSrcError(err=InstSrcError.E_UNDEF)

            if type(options.source) is vmdisks.Disk:
                vdev = options.source.vdev
                if len(filter(lambda d: d.vdev==vdev and d.exists(), options.disks)) == 0:
                    log.debug("validation failed: No matching disk found for PV install source")
                    raise InstSrcError(InstSrcError.E_NOT_FOUND)

        if options.graphics not in ('none', 'para'):
            raise ParamError(err=ParamError.E_INVALID, details='graphics')

