# Functions to enable host devices within a VM
#
# Copyright 2012 Novell, Inc.
# Author:
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 

import caps
import util
from VMDefaults import VMDefaults
from exceptions import *

#-------------------------------------------------------------------------------
# Function: parse_string
# Parms:
#   host_device:    This is a comma separated string that contains the information
#           for the pci or usb device to be passed to the VM
#           The format for the string is:
#               pci=[DOMAIN:BUS:SLOT.FUNC[MANAGED]], usb=[BUS:DEVICE:VENDOR_ID:PRODUCT_ID[MANAGED]]
#               MANAGED is optional and is set to allow libvirt to manage the device
#
# Invalid entries in the host device parameter will cause an exception.
# The input string is parsed and a HostDevice (class) is created.
#-------------------------------------------------------------------------------
def parse_string(host_device):
    if host_device == "":
        raise ParamError(err=ParamError.E_INVALID, details=host_device)
    params = host_device.split(',')
    type = None
    host_devices = []
    i = 0
    for p in params:
        managed = False
        if p.startswith('pci='):
            type = 'pci'
        elif p.startswith('usb='):
            type = 'usb'
        else:
            raise ParamError(err=ParamError.E_INVALID, details=p)
        parts = p[4:].split(':')
        len_parts = len(parts)
        if type == 'pci' and len_parts == 4 or \
           type == 'usb' and len_parts == 5:
            len_parts = len_parts - 1
            if parts[len_parts] is not '0':
                managed = True
        if type == 'pci' and len_parts != 3 or \
           type == 'usb' and len_parts != 4:
            raise ParamError(err=ParamError.E_INVALID, details=parts)
        if type is 'pci':
            subparts = parts[2].split('.')
            if len(subparts) != 2:
                raise ParamError(err=ParamError.E_INVALID, details=subparts)
            dev = PCIHostDevice(None, parts[0], parts[1], subparts[0], subparts[1])
        else:
            dev = USBHostDevice(None, parts[0], parts[1], parts[2], parts[3])
        i = i + 1
        hd = HostDevice(i, type, managed, dev)
        if host_devices:
            host_devices.append(hd)
        else:
            host_devices = [hd]
    return host_devices

# Create the list of PCI devices
def get_host_pci_devices():
    lines = util.run(['/sbin/lspci', '-D'])
    i = 0
    pci_devices = []
    for line in lines.splitlines():
        pci_devices.append(line)
    return pci_devices

# Initialize a ListStore with the list of USB devices
def init_host_pci_devices_list(list):
    pci_devices = get_host_pci_devices()
    i = 0
    for pci in pci_devices:
        list.append(pci)

# Create the list of USB devices
def get_host_usb_devices():
    lines = util.run(['/usr/bin/lsusb'])
    i = 0
    usb_devices = []
    for line in lines.splitlines():
        usb_devices.append(line)
    return usb_devices

# Initialize a ListStore with the list of USB devices
def init_host_usb_devices_list(list):
    usb_devices = get_host_usb_devices()
    i = 0
    for usb in usb_devices:
        list.append(usb)

class PCIHostDevice(object):
    def __init__(self, full_description=None,domain=None,bus=None,slot=None,function=None):
        # Example lspci -D
        # 0000:00:19.0 Ethernet controller: Intel Corporation 82566DC Gigabit Network Connection (rev 02)
        #   domain = 0000
        #   bus = 00
        #   slot = 19
        #   function = 0
        #   Description = Ethernet controller: Intel Corporation 82566DC Gigabit Network Connection (rev 02)
        if full_description:
            self.full_description = full_description
            parts=full_description.split(':')
            self.domain=parts[0]
            self.bus=parts[1]
            subparts=parts[2].split('.')
            self.slot = subparts[0]
            self.function = subparts[1].split(' ')[0]
        else:
            self.domain = domain.rjust(4,'0')
            self.bus = bus.rjust(2,'0')
            self.slot = slot.rjust(2,'0')
            self.function = function
            self.full_description = self.lookup_pci_device()
    def copy(self):
        return PCIHostDevice(self.full_description)
    def lookup_pci_device(self):
        lines = util.run(['/sbin/lspci', '-D'])
        i = 0
        look_for = "%s:%s:%s.%s" % (self.domain, self.bus, self.slot, self.function)
        for line in lines.splitlines():
            parts = line.split(' ',1)
            if parts[0] == look_for:
                return line
            i = i + 1
        raise ParamError(err=ParamError.E_INVALID, details=look_for)

class USBHostDevice(object):
    def __init__(self, full_description=None,bus=None,device=None,vendor_id=None,product_id=None):
        # Example lsusb
        # Bus 003 Device 002: ID 413c:2005 Dell Computer Corp. RT7D50 Keyboard
        #   bus = 001
        #   device = 001
        #   vendor ID = 1d6b
        #   produce ID = 0002
        #   description = Dell Computer Corp. RT7D50 Keyboard
        if full_description:
            self.full_description = full_description
            parts=full_description.split(' ')
            self.bus = parts[1]
            self.device = parts[3].rstrip(':')
            subparts = parts[5].split(':')
            self.vendor_id = subparts[0]
            self.product_id = subparts[1]
        else:
            self.bus = bus.rjust(3,'0')
            self.device = device.rjust(3,'0')
            self.vendor_id = vendor_id.rjust(4,'0')
            self.product_id = product_id.rjust(4,'0')
            self.full_description = self.lookup_usb_device()
    def copy(self):
        return USBHostDevice(self.full_description)
    def lookup_usb_device(self):
        lines = util.run(['/usr/sbin/lsusb'])
        i = 0
        look_for = "Bus %s Device %s: ID %s:%s" % (self.bus, self.device, self.vendor_id, self.product_id)
        for line in lines.splitlines():
            if line.startswith(look_for):
                return line
            i = i + 1
        raise ParamError(err=ParamError.E_INVALID, details=look_for)

class HostDevice(object):
    def __init__(self, number=None, type=None, managed=False, host_device=None):
        if type not in ('pci', 'usb') or host_device is None:
            raise ParamError(err=ParamError.E_INVALID, details='type is %s' % type)
        self.number = number
        self.type = type
        self.managed = managed
        self.host_device = host_device
        self.description = host_device.full_description

    def copy(self):
        if self.type is 'pci':
            return PCIHostDevice(full_description=self.host_device.full_description)
        else:
            return USBHostDevice(full_description=self.host_device.full_description)

