################################################################################
# Initially derived from virt-inst, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 
################################################################################

from   console import Console
from   log     import log
import threading
import time

try:
    from xen.xend.xenstore.xstransact import xstransact
except:
    pass
try:
    from xen.xend.xenstore.xswatch import xswatch
except:
    pass


################################################################################
# CLASS: XenConsole
#
# This class contains the Xen console implementation.
#
################################################################################
class XenConsole(Console):

    #---------------------------------------------------------------------------
    # Class:  XenConsole
    # Method: get_domid_from_uuid
    # Parms:
    #   uuid:   uuid from the domain
    #
    # This method gets the domain ID for the given UUID.
    #
    #---------------------------------------------------------------------------
    def get_domid_from_uuid(self, uuid):
        vm_path = '/vm/%s' % uuid
        t = xstransact(path='/local/domain/')
        domids = t.list()
        for domid in domids:
            path = t.read('%s/vm' % domid)
            if path == vm_path:
                d = t.read('%s/domid' % domid)
                if d == 'None':
                    continue
                domid = int(domid)
                log.debug("Domain %s has ID %d" % (uuid, domid))
                t.abort()
                return domid
        log.debug("Domain %s is not running" % uuid)
        return None


    #---------------------------------------------------------------------------
    # Class:  XenConsole
    # Method: get_tty_from_domid
    # Parms:
    #   domid:  domain ID
    #
    # This method gets the tty number for a given domain.
    #
    # This method is stubbed out in the base class.  It is expected that
    # anyone that subclasses this class will create their own method.
    #---------------------------------------------------------------------------
    def get_tty_from_domid(self, domid):
        if domid is None:
            return None
        tty = None
        console_path = '/local/domain/%d/console/tty' % domid
        tty = xstransact.Read(console_path)
        log.info('Domain ID %d has tty %s' % (domid, str(tty)))
        return tty


    #---------------------------------------------------------------------------
    # Class:  XenConsole
    # Method: wait_for_domid_to_exit
    # Parms:
    #   domid:  The domain ID of the domain that we are waiting to exit.
    #   uuid:   UUID from the domain
    #
    # This method waits for the given domain to exit.
    #
    #---------------------------------------------------------------------------
    def wait_for_domid_to_exit(self, domid, uuid):
        """Wait for domid to exit (or, for domid's UUID to not match
        the specified uuid, which would indicate a race condition in
        which domid has already exited and was replace by another)."""
        if domid is None:
            return
        uuid_path = '/local/domain/%d/vm' % domid
        log.debug("Watching xenstore path '%s'" % uuid_path)
        cv = threading.Condition()
        def fn(path):
            log.debug("Watch fired for '%s'" % uuid_path)
            cv.acquire()
            u = xstransact.Read(uuid_path)
            log.debug("UUID is '%s'", u)
            if u is None:
                again = False
            else:
                u = u.split('/')[-1]  # Why does uuid have "/vm/" prefix?
                if u == uuid:
                    again = True
                else:
                    again = False
            if not again:
                cv.notifyAll()
            cv.release()
            return again
        cv.acquire()
        w = xswatch(uuid_path, fn)
        cv.wait()
        cv.release()


    #---------------------------------------------------------------------------
    # Class:  XenConsole
    # Method: get_vnc_port_from_domid
    #
    # Parms:
    #   domid:      The domain ID of the domain that we are waiting to exit.
    #   timeout:    Wait only so long.
    #
    # This method finds the vnc port for the given domain.
    #
    # This method is stubbed out in the base class.  It is expected that
    # anyone that subclasses this class will create their own method.
    #---------------------------------------------------------------------------
    def get_vnc_port_from_domid(self, domid, timeout=0.0):
        if domid is None:
            return None
        ports = [None]
        console_path = '/local/domain/%d/console/vnc-port' % domid
        log.debug("Watching xenstore path '%s'" % console_path)
        cv = threading.Condition()
        def fn(path):
            log.debug("Watch fired for '%s'" % console_path)
            again = True
            cv.acquire()
            p = xstransact.Read(console_path)
            if p is not None:
                log.debug("vnc-port is '%s'", p)
                ports[0] = p
                cv.notifyAll()
                again = False
            cv.release()
            return again
        cv.acquire()
        w = xswatch(console_path, fn)
        cv.wait(timeout)
        cv.release()
        port = ports[0]
        log.info('Domain ID %d has vnc port %s' % (domid, port))
        if port is not None:
            port = int(port)
        return port

