################################################################################
# Communicates with xend
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 
#
# This is the file that communicates with the xen hypervisor.
#
################################################################################

import time
import os
import re
from hypervisor import Hypervisor
from exceptions import ParamError, XenError
import util
import caps
from log import log

def parse_string(uri):
    if not uri.startswith("xen") and not uri.startswith("qemu"):
        raise ParamError(err=ParamError.E_INVALID, details=uri)
    return Xl(uri)

################################################################################
# Class: XenHypervisor
#
# Xen Hypervisor specifics.
################################################################################
class Xl(Hypervisor):
    """Thin abstraction to talk to Xl.

    In case you don't want to use libvirt.
    """

    #---------------------------------------------------------------------------
    # Method: __init__
    #
    # Comments:
    #   Sets the connection information to talk to xen.
    #
    #---------------------------------------------------------------------------
    def __init__(self, uri):
        if os.getuid() != 0:
            raise XenError(err=XenError.E_ROOT, details=uri)
        self.use_libvirt = False

    #---------------------------------------------------------------------------
    # Method: Check if string is a valid uuid
    #---------------------------------------------------------------------------
    def _valid_uuid(self, uuid):
        regex = re.compile('^[a-f0-9]{8}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[a-f0-9]{12}\Z', re.I)
        match = regex.match(uuid)
        return bool(match)

    #---------------------------------------------------------------------------
    # Method: Wait for a shutdown to be successful
    #---------------------------------------------------------------------------
    def _wait_for_shutdown(self, vmname, timeout):
        for _ in xrange(timeout):
            if not self.domain_running(vmname):
                return True
            time.sleep(1.0)
        return False

    #---------------------------------------------------------------------------
    # Method: Wait for a shutdown to be successful
    #---------------------------------------------------------------------------
    def _get_vm_name(self, domid):
        vmname = str(domid)
        if self._valid_uuid(domid):
            uuid_path = '/vm/' + str(domid) + '/name'
            try:
                output = util.run(['/usr/bin/xenstore-read', '-s', uuid_path])
                vmname = output.rstrip()
            except:
                pass
        return vmname

    #---------------------------------------------------------------------------
    # Method: delete
    #---------------------------------------------------------------------------
    def delete(self, domid):
        """Deletes a running instance using xl; erases"""
        vmname = self._get_vm_name(domid)
        try:
            output = util.run(['/usr/sbin/xl', 'shutdown', vmname])
            if self._wait_for_shutdown(vmname, 20):
                return
            log.debug("shutdown not working, use destroy")
            util.run(['/usr/sbin/xl', 'destroy', vmname])
        except:
            log.debug("xl delete failed for %s" % vmname)
            pass

    #---------------------------------------------------------------------------
    # Method: destroy
    #---------------------------------------------------------------------------
    def destroy(self, domid):
        """Destroys a running instance using xl; does not erase"""
        if not self.domain_running(domid):
            return
        vmname = self._get_vm_name(domid)
        try:
            output = util.run(['/usr/sbin/xl', 'shutdown', vmname])
            if self._wait_for_shutdown(vmname, 20):
                return
            log.debug("shutdown not working, use destroy")
            output = util.run(['/usr/sbin/xl', 'destroy', str(domid)])
        except:
            pass

    #---------------------------------------------------------------------------
    # Method: domain
    #---------------------------------------------------------------------------
    def domain(self, domid):
        # xl doesn't have managed domains
        return self.domain_running(domid)

    #---------------------------------------------------------------------------
    # Method: get_dom0_min_memory
    #---------------------------------------------------------------------------
    def get_dom0_min_memory(self):
        return 0

    #---------------------------------------------------------------------------
    # Method: domain_running
    #---------------------------------------------------------------------------
    def domain_running(self, domid):
        try:
            vmname_path = str(domid)
            output = util.run(['/usr/sbin/xl', 'list', vmname_path])
            return True
        except:
            # 'xl list' will fail on uuid or domain ID.  Try other methods
            if self._valid_uuid(domid):
                try:
                    uuid_path = '/vm/' + str(domid)
                    output = util.run(['/usr/bin/xenstore-read', '-s', uuid_path])
                    return True
                except:
                    pass
            try:
                domid_path = '/local/domain/' + str(domid)
                output = util.run(['/usr/bin/xenstore-read', '-s', domid_path])
                return True
            except:
                return False

    #---------------------------------------------------------------------------
    # Method: info
    #---------------------------------------------------------------------------
    def info(self):
        maxmem = caps.free_memory(True)
        num_cpus = caps.cpu_processors()
        total_memory = caps.maximum_memory()
        return dict([('nr_cpus', num_cpus), ('total_memory', total_memory), ('max_hvm_memory', maxmem), ('max_para_memory', maxmem)])

    #---------------------------------------------------------------------------
    # Method: lookup
    #---------------------------------------------------------------------------
    def lookup(self, domid):
        """Lookup the domain by name, UUID, or hypervisor ID"""
        return self.domain_running(domid)

    #---------------------------------------------------------------------------
    # Method: wait_domain
    #---------------------------------------------------------------------------
    def wait_domain(self, domid):
        #  This is ugly (polling instead of event driven) but XenAPI doesn't
        #  yet grok unix socket.
        #
        #  If domain and 'state' are both found, then assume VM is trying
        #  to come up, so keep waiting.  Otherwise, fail.
        for _ in xrange(10):
            wait = False
            if self.domain_running(domid):
                return True
            time.sleep(1.0)
        return False

