# libvirt-style XML parsing
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 

import libxml2

from exceptions import *
from log import log
from vmdisks.disks import Disk
from nics import XenNic
from nics import KvmNic
import caps

def tryElem(ctxt, xpath):
    e = ctxt.xpathEval(xpath)
    if len(e):
        return e[0]
    return None

def tryElems(ctxt, xpath):
    return ctxt.xpathEval(xpath)

def reqElem(ctxt, xpath):
    e = tryElem(ctxt, xpath)
    if e is None:
        raise ParamError(err=ParamError.E_INVALID, details=msg.malformed_xml % {'tag':xpath})
    return e

def tryStr(ctxt, xpath):
    e = tryElem(ctxt, xpath)
    if e is None:
        return None
    return e.content

def reqStr(ctxt, xpath):
    return reqElem(ctxt, xpath).content

def tryInt(ctxt, xpath):
    e = tryElem(ctxt, xpath)
    if e is None:
        return None
    return int(e.content)

def reqInt(ctxt, xpath):
    e = reqElem(ctxt, xpath)
    return int(e.content)

def tryExists(ctxt, xpath):
    return (tryElem(ctxt, xpath) is not None)

def K2M(k):
    if k is None:
        return None
    return k/1024

def set(options, name, value):
    if value is not None:
        setattr(options, name, value)

def get_disks(devices_xml):
    disks = []
    for disk in tryElems(devices_xml, 'disk'):
        pdev = tryElem(disk, 'source/@dev') or reqElem(disk, 'source/@file')
        pdev = pdev.content
        vdev = reqElem(disk, 'target/@dev').content
        bus = tryElem(disk, 'target/@bus')
        if bus:
            bus = bus.content
            if bus != 'virtio':
                bus = None
        vdevType = reqElem(disk, '@device').content
        ro = (tryElem(disk, 'readonly') is not None)
        disks.append(Disk(pdev=pdev, vdev=vdev, vdevType=vdevType, ro=ro, virtio=bus))
    return disks

def merge(xml_file, options):
    """Extract settings from an XML file and overlay them onto 'options'."""
    log.info("Loading XML settings from %s" % (xml_file,))

    #FIXME: I believe that this is xen-specific and will not work for kvm.
    #       However, this is being called from vm-install and vm-install
    #       has the capability to determine if we're xen or kvm.

    try:
        doc = libxml2.parseFile(xml_file)
    except:
        raise ParamError(err=ParamError.E_INVALID, details=msg.parse_failed % {'file':xml_file})
    try:
        ctxt = doc.xpathNewContext()
        dom = reqElem(ctxt, "/domain[@type='xen' or @type='kvm']")
        set(options, 'name', tryStr(dom, 'name'))
        set(options, 'description', tryStr(dom, 'description'))
        set(options, 'memoryMB', K2M(tryInt(dom, 'currentMemory')))
        set(options, 'max_memoryMB', K2M(tryInt(dom, 'memory')))
        set(options, 'vcpus', tryInt(dom, 'vcpu'))
        set(options, 'uuid', tryStr(dom, 'uuid'))
        set(options, 'on_crash', tryStr(dom, 'on_crash'))
        set(options, 'on_poweroff', tryStr(dom, 'on_poweroff'))
        set(options, 'on_reboot', tryStr(dom, 'on_reboot'))
        set(options, 'keymap', tryStr(dom, 'keymap'))
        #set(options, 'loader', tryStr(dom, 'bootloader'))
        os = tryElem(dom, 'os')
        if os:
            os_source = None
            os_data = str(os)
            os_data = os_data.split(" ")
            for data in os_data:
                if data.startswith("type="):
                    data = data.split('"')
                    if len(data) > 1:
                        set(options, 'os_type', data[1])
                elif data.startswith("source="):
                    data = data.split('"')
                    if len(data) > 1:
                        os_source = data[1]
                        set(options, 'source', os_source)
                elif data.startswith("install=") or data.startswith("<cmdline>install="):
                    if os_source is None:
                        data = data.split('=')
                        if len(data) > 1:
                            os_install = data[1].split("</cmdline>")
                            os_source = os_install[0]
                            set(options, 'source', os_source)
            full_virt = tryStr(os, 'type')
            if full_virt is not None:
                set(options, 'full_virt', full_virt == 'hvm')
                if options.full_virt:
                    set(options, 'loader', tryStr(os, 'loader'))
                    #boot = tryElem(os, 'boot')
                else:
                    set(options, 'loader', tryStr(os, 'bootloader'))
                    if options.loader:
                        # bootargs are regenerated post-install.
                        pass
                    else:
                        set(options, 'kernel', tryStr(os, 'kernel'))
                        set(options, 'initrd', tryStr(os, 'initrd'))
                    set(options, 'extra_args', tryStr(os, 'cmdline'))
        if options.full_virt:
            features = tryElem(dom, 'features')
            if features:
                options.apic = tryExists(features, 'apic')
                options.acpi = tryExists(features, 'acpi')
                options.pae = tryExists(features, 'pae')
        devices = tryElem(dom, 'devices')
        if devices:
            options.disks = get_disks(devices)
            if options.nics is None:
                options.nics = []
            for nic in tryElems(devices, 'interface'):
                mac = tryElem(nic, 'mac/@address')
                if mac: mac = mac.content
                bridge = tryElem(nic, 'source/@bridge')
                if bridge: bridge = bridge.content
                model = tryElem(nic, '@model')
                if model: model = model.content
                options.nics.append(XenNic(mac=mac, bridge=bridge, model=model,
                                           full_virt=options.full_virt))
            graphics = tryElem(devices, 'graphics')
            # FIXME
    finally:
        # libxml2-python requires an explicit free of the doc!
        doc.freeDoc()
